import shutil

import pytest
from distutils.dir_util import copy_tree

import scvi
from tests.data.utils import generic_setup_adata_manager


def pytest_addoption(parser):
    """Docstring for pytest_addoption."""
    parser.addoption(
        "--internet-tests",
        action="store_true",
        default=False,
        help="Run tests that retrieve stuff from the internet. This increases test time.",
    )
    parser.addoption(
        "--multigpu-tests",
        action="store_true",
        default=False,
        help="Run tests that are desinged for multiGPU.",
    )
    parser.addoption(
        "--optional",
        action="store_true",
        default=False,
        help="Run tests that are optional.",
    )
    parser.addoption(
        "--accelerator",
        action="store",
        default="cpu",
        help="Option to specify which accelerator to use for tests.",
    )
    parser.addoption(
        "--devices",
        action="store",
        default="auto",
        help="Option to specify which devices to use for tests.",
    )
    parser.addoption(
        "--seed",
        action="store",
        default=0,
        help="Option to specify which scvi-tools seed to use for tests.",
    )
    parser.addoption(
        "--private",
        action="store_true",
        default=False,
        help="Run tests that are private.",
    )


def pytest_configure(config):
    """Docstring for pytest_configure."""
    config.addinivalue_line("markers", "optional: mark test as optional.")


def pytest_collection_modifyitems(config, items):
    """Docstring for pytest_collection_modifyitems."""
    run_internet = config.getoption("--internet-tests")
    skip_internet = pytest.mark.skip(reason="need --internet-tests option to run")
    skip_non_internet = pytest.mark.skip(reason="test not having a pytest.mark.internet decorator")
    for item in items:
        # All tests marked with `pytest.mark.internet` get skipped unless
        # `--internet-tests` passed
        if not run_internet and ("internet" in item.keywords):
            item.add_marker(skip_internet)
        # Skip all tests not marked with `pytest.mark.internet` if `--internet-tests` passed
        elif run_internet and ("internet" not in item.keywords):
            item.add_marker(skip_non_internet)

    run_optional = config.getoption("--optional")
    skip_optional = pytest.mark.skip(reason="need --optional option to run")
    skip_non_optional = pytest.mark.skip(reason="test not having a pytest.mark.optional decorator")
    for item in items:
        # All tests marked with `pytest.mark.optional` get skipped unless
        # `--optional` passed
        if not run_optional and ("optional" in item.keywords):
            item.add_marker(skip_optional)
        # Skip all tests not marked with `pytest.mark.optional` if `--optional` passed
        elif run_optional and ("optional" not in item.keywords):
            item.add_marker(skip_non_optional)

    run_private = config.getoption("--private")
    skip_private = pytest.mark.skip(reason="need --private option to run")
    skip_non_private = pytest.mark.skip(reason="test not having a pytest.mark.private decorator")
    for item in items:
        # All tests marked with `pytest.mark.private` get skipped unless
        # `--private` passed
        if not run_private and ("private" in item.keywords):
            item.add_marker(skip_private)
        # Skip all tests not marked with `pytest.mark.private` if `--private` passed
        elif run_private and ("private" not in item.keywords):
            item.add_marker(skip_non_private)

    run_multigpu = config.getoption("--multigpu-tests")
    skip_multigpu = pytest.mark.skip(reason="need --multigpu-tests option to run")
    skip_non_multigpu = pytest.mark.skip(reason="test not having a pytest.mark.multigpu decorator")
    for item in items:
        # All tests marked with `pytest.mark.multigpu` get skipped unless
        # `--multigpu-tests` passed
        if not run_multigpu and ("multigpu" in item.keywords):
            item.add_marker(skip_multigpu)
        # Skip all tests not marked with `pytest.mark.multigpu` if `--multigpu-tests` passed
        elif run_multigpu and ("multigpu" not in item.keywords):
            item.add_marker(skip_non_multigpu)


@pytest.fixture(scope="session")
def save_path(tmp_path_factory):
    """Docstring for save_path."""
    dir = tmp_path_factory.mktemp("temp_data", numbered=False)
    path = str(dir)
    copy_tree("tests/test_data", path)
    yield path + "/"
    shutil.rmtree(str(tmp_path_factory.getbasetemp()))


@pytest.fixture(scope="session")
def accelerator(request):
    """Docstring for accelerator."""
    return request.config.getoption("--accelerator")


@pytest.fixture(scope="session")
def devices(request):
    """Docstring for devices."""
    return request.config.getoption("--devices")


@pytest.fixture(scope="session")
def mock_contrastive_adata():
    """Synthetic contrastive adata."""
    adata = scvi.data.synthetic_iid(n_batches=2)
    adata = adata[:-3, :]  # Unequal technical batch sizes.
    adata.layers["raw_counts"] = adata.X.copy()
    return adata


@pytest.fixture(scope="session")
def mock_contrastive_adata_manager(mock_contrastive_adata):
    """Anndata manager for synthetic contrastive data."""
    return generic_setup_adata_manager(
        adata=mock_contrastive_adata,
        batch_key="batch",
        labels_key="labels",
        layer="raw_counts",
    )


@pytest.fixture(scope="session")
def mock_background_indices(mock_contrastive_adata):
    """Indices for background data in ``mock_contrastive_adata_manager``."""
    adata = mock_contrastive_adata
    return adata.obs.index[(adata.obs["batch"] == "batch_0")].astype(int).tolist()


@pytest.fixture(scope="session")
def mock_target_indices(mock_contrastive_adata):
    """Indices for target data in ``mock_contrastive_adata_manager``."""
    adata = mock_contrastive_adata
    return adata.obs.index[(adata.obs["batch"] == "batch_1")].astype(int).tolist()


@pytest.fixture(autouse=True)
def set_seed(request):
    """Sets the seed for each test."""
    from scvi import settings

    settings.seed = int(request.config.getoption("--seed"))
    return None



import copy
import logging

import numpy as np
from torch.utils.data import (
    BatchSampler,
    DataLoader,
    RandomSampler,
    Sampler,
    SequentialSampler,
)

from scvi import settings
from scvi.data import AnnDataManager

from ._samplers import BatchDistributedSampler

logger = logging.getLogger(__name__)


class AnnDataLoader(DataLoader):
    """DataLoader for loading tensors from AnnData objects.

    Parameters
    ----------
    adata_manager
        :class:`~scvi.data.AnnDataManager` object with a registered AnnData object.
    indices
        The indices of the observations in `adata_manager.adata` to load.
    batch_size
        Minibatch size to load each iteration. If `distributed_sampler` is `True`,
        refers to the minibatch size per replica. Thus, the effective minibatch
        size is `batch_size` * `num_replicas`.
    shuffle
        Whether the dataset should be shuffled prior to sampling.
    sampler
        Defines the strategy to draw samples from the dataset. Can be any Iterable with __len__
        implemented. If specified, shuffle must not be specified. By default, we use a custom
        sampler that is designed to get a minibatch of data with one call to __getitem__.
    drop_last
        If `True` and the dataset is not evenly divisible by `batch_size`, the last
        incomplete batch is dropped. If `False` and the dataset is not evenly divisible
        by `batch_size`, then the last batch will be smaller than `batch_size`.
    drop_dataset_tail
        Only used if `distributed_sampler` is `True`. If `True` the sampler will drop
        the tail of the dataset to make it evenly divisible by the number of replicas.
        If `False`, then the sampler will add extra indices to make the dataset evenly
        divisible by the number of replicas.
    data_and_attributes
        Dictionary with keys representing keys in data registry (``adata_manager.data_registry``)
        and value equal to desired numpy loading type (later made into torch tensor) or list of
        such keys. A list can be used to subset to certain keys in the event that more tensors than
        needed have been registered. If ``None``, defaults to all registered data.
    iter_ndarray
        Whether to iterate over numpy arrays instead of torch tensors
    distributed_sampler
        ``EXPERIMENTAL`` Whether to use :class:`~scvi.dataloaders.BatchDistributedSampler` as the
        sampler. If `True`, `sampler` must be `None`.
    load_sparse_tensor
        ``EXPERIMENTAL`` If ``True``, loads data with sparse CSR or CSC layout as a
        :class:`~torch.Tensor` with the same layout. Can lead to speedups in data transfers to
        GPUs, depending on the sparsity of the data.
    **kwargs
        Additional keyword arguments passed into :class:`~torch.utils.data.DataLoader`.

    Notes
    -----
    If `sampler` is not specified, a :class:`~torch.utils.data.BatchSampler` instance is
    passed in as the sampler, which retrieves a minibatch of data with one call to
    :meth:`~scvi.data.AnnTorchDataset.__getitem__`. This is useful for fast access to
    sparse matrices as retrieving single observations and then collating is inefficient.
    """

    def __init__(
        self,
        adata_manager: AnnDataManager,
        indices: list[int] | list[bool] | None = None,
        batch_size: int = 128,
        shuffle: bool = False,
        sampler: Sampler | None = None,
        drop_last: bool = False,
        drop_dataset_tail: bool = False,
        data_and_attributes: list[str] | dict[str, np.dtype] | None = None,
        iter_ndarray: bool = False,
        distributed_sampler: bool = False,
        load_sparse_tensor: bool = False,
        **kwargs,
    ):
        if indices is None:
            indices = np.arange(adata_manager.adata.shape[0])
        else:
            if hasattr(indices, "dtype") and indices.dtype is np.dtype("bool"):
                indices = np.where(indices)[0].ravel()
            indices = np.asarray(indices)
        self.indices = indices
        self.dataset = adata_manager.create_torch_dataset(
            indices=indices,
            data_and_attributes=data_and_attributes,
            load_sparse_tensor=load_sparse_tensor,
        )
        if "num_workers" not in kwargs:
            kwargs["num_workers"] = settings.dl_num_workers
        if "persistent_workers" not in kwargs:
            kwargs["persistent_workers"] = settings.dl_persistent_workers

        self.kwargs = copy.deepcopy(kwargs)

        if sampler is not None and distributed_sampler:
            raise ValueError("Cannot specify both `sampler` and `distributed_sampler`.")

        # custom sampler for efficient minibatching on sparse matrices
        if sampler is None:
            if not distributed_sampler:
                sampler_cls = SequentialSampler if not shuffle else RandomSampler
                sampler = BatchSampler(
                    sampler=sampler_cls(self.dataset),
                    batch_size=batch_size,
                    drop_last=drop_last,
                )
            else:
                sampler = BatchDistributedSampler(
                    self.dataset,
                    batch_size=batch_size,
                    drop_last=drop_last,
                    drop_dataset_tail=drop_dataset_tail,
                    shuffle=shuffle,
                )
            # do not touch batch size here, sampler gives batched indices
            # This disables PyTorch automatic batching, which is necessary
            # for fast access to sparse matrices
            self.kwargs.update({"batch_size": None, "shuffle": False})

        self.kwargs.update({"sampler": sampler})

        if iter_ndarray:
            self.kwargs.update({"collate_fn": lambda x: x})

        super().__init__(self.dataset, **self.kwargs)



from scvi.hub._url import validate_url


def test_validate_url():
    valid = "https://scvi-tools.org/"
    assert validate_url(valid)

    invalid = "scvi-tools.org"
    assert not validate_url(invalid)



from collections.abc import Iterable, Sequence

import numpy as np
import pyro
import pyro.distributions as dist
import pyro.poutine as poutine
import torch
import torch.nn as nn
import torch.utils.data
from torch.distributions import constraints
from torch.nn.functional import softmax, softplus

from scvi._constants import REGISTRY_KEYS
from scvi.module.base import PyroBaseModuleClass, auto_move_data

from ._components import ConditionalDenseNN

TEST_BETA = 1.0


class DecipherPyroModule(PyroBaseModuleClass):
    """Pyro Module for the Decipher model.

    This module implements the Decipher model for dimensionality reduction and
    interpretable representation learning in single-cell RNA sequencing data.

    Parameters
    ----------
    dim_genes
        Number of genes (features) in the dataset.
    dim_v
        Dimension of the interpretable latent space v.
    dim_z
        Dimension of the intermediate latent space z.
    layers_v_to_z
        Hidden layer sizes for the v to z decoder network.
    layers_z_to_x
        Hidden layer sizes for the z to x decoder network.
    beta
        Regularization parameter for the KL divergence.
    """

    def __init__(
        self,
        dim_genes: int,
        dim_v: int = 2,
        dim_z: int = 10,
        layers_v_to_z: Sequence[int] = (64,),
        layers_z_to_x: Sequence[int] = (),
        beta: float = 0.1,
    ):
        super().__init__()
        self.dim_v = dim_v
        self.dim_z = dim_z
        self.dim_genes = dim_genes
        self.layers_v_to_z = layers_v_to_z
        self.layers_z_to_x = layers_z_to_x
        self.beta = beta

        self.decoder_v_to_z = ConditionalDenseNN(dim_v, layers_v_to_z, [dim_z] * 2)
        self.decoder_z_to_x = ConditionalDenseNN(dim_z, layers_z_to_x, [dim_genes])
        self.encoder_x_to_z = ConditionalDenseNN(dim_genes, [128], [dim_z] * 2)
        self.encoder_zx_to_v = ConditionalDenseNN(
            dim_genes + dim_z,
            [128],
            [dim_v] * 2,
        )

        self.theta = None

        self._epsilon = 1e-5

        self.n_obs = None  # Populated by PyroTrainingPlan

        # Hack: to allow auto_move_data to infer device.
        self._dummy_param = nn.Parameter(torch.empty(0))

    @property
    def device(self):
        return self._dummy_param.device

    @staticmethod
    def _get_fn_args_from_batch(tensor_dict: dict[str, torch.Tensor]) -> Iterable | dict:
        x = tensor_dict[REGISTRY_KEYS.X_KEY]
        return (x,), {}

    @auto_move_data
    def model(self, x: torch.Tensor, beta: float | None = None):
        pyro.module("decipher", self)

        self.theta = pyro.param(
            "theta",
            x.new_ones(self.dim_genes),
            constraint=constraints.positive,
        )

        with (
            pyro.plate("batch", x.shape[0]),
            poutine.scale(scale=1.0),
        ):
            with poutine.scale(scale=beta or self.beta):
                prior = dist.Normal(0, x.new_ones(self.dim_v)).to_event(1)
                v = pyro.sample("v", prior)

            z_loc, z_scale = self.decoder_v_to_z(v)
            z_scale = softplus(z_scale)
            z = pyro.sample("z", dist.Normal(z_loc, z_scale).to_event(1))

            mu = self.decoder_z_to_x(z)
            mu = softmax(mu, dim=-1)
            library_size = x.sum(axis=-1, keepdim=True)
            # Parametrization of Negative Binomial by the mean and inverse dispersion
            # See https://github.com/pytorch/pytorch/issues/42449
            # noinspection PyTypeChecker
            logit = torch.log(library_size * mu + self._epsilon) - torch.log(
                self.theta + self._epsilon
            )
            # noinspection PyUnresolvedReferences
            if self.device.type == "mps":
                # TODO: TORCH MPS FIX
                x_dist = dist.NegativeBinomial(
                    total_count=self.theta.contiguous() + self._epsilon, logits=logit.contiguous()
                )
            else:
                x_dist = dist.NegativeBinomial(
                    total_count=self.theta + self._epsilon, logits=logit
                )
            pyro.sample("x", x_dist.to_event(1), obs=x)

    @auto_move_data
    def guide(self, x: torch.Tensor, beta: float | None = None):
        pyro.module("decipher", self)
        with (
            pyro.plate("batch", x.shape[0]),
            poutine.scale(scale=1.0),
        ):
            x = torch.log1p(x)

            z_loc, z_scale = self.encoder_x_to_z(x)
            z_scale = softplus(z_scale)
            posterior_z = dist.Normal(z_loc, z_scale).to_event(1)
            z = pyro.sample("z", posterior_z)

            zx = torch.cat([z, x], dim=-1)
            v_loc, v_scale = self.encoder_zx_to_v(zx)
            v_scale = softplus(v_scale)
            with poutine.scale(scale=beta or self.beta):
                posterior_v = dist.Normal(v_loc, v_scale).to_event(1)
                pyro.sample("v", posterior_v)
        return z_loc, v_loc, z_scale, v_scale

    def predictive_log_likelihood(self, x: torch.Tensor, n_samples=5):
        """
        Calculate the predictive log-likelihood for a Decipher module.

        This function performs multiple runs through the dataloader to obtain
        an empirical estimate of the predictive log-likelihood. It calculates the
        log-likelihood for each run and returns the average. The beta parameter
        of the Decipher module is temporarily modified and restored even if an
        exception occurs. Used by default as an early stopping criterion.

        Parameters
        ----------
        x : torch.Tensor
            Batch of data to compute the log-likelihood for.
        n_samples : int, optional
            Number of passes through the dataloader (default is 5).

        Returns
        -------
        float
            The average estimated predictive log-likelihood across multiple runs.
        """
        log_weights = []
        for _ in range(n_samples):
            guide_trace = poutine.trace(self.guide).get_trace(x, beta=TEST_BETA)
            model_trace = poutine.trace(poutine.replay(self.model, trace=guide_trace)).get_trace(
                x, beta=TEST_BETA
            )
            log_weights.append(model_trace.log_prob_sum() - guide_trace.log_prob_sum())

        log_z = torch.logsumexp(torch.tensor(log_weights) - np.log(n_samples), 0)
        return log_z.item()



from __future__ import annotations

from os.path import join

import pytest
import torch

from scvi.nn import Embedding


@pytest.mark.parametrize("num_embeddings", [10])
@pytest.mark.parametrize("embedding_dim", [5])
@pytest.mark.parametrize("init", [2, [0, 1]])
@pytest.mark.parametrize("freeze_prev", [True, False])
def test_embedding_extend(
    num_embeddings: int,
    embedding_dim: int,
    init: int | list[int],
    freeze_prev: bool,
):
    embedding = Embedding(num_embeddings, embedding_dim)
    ext_embedding = Embedding.extend(embedding, init=init, freeze_prev=freeze_prev)
    n_init = len(init) if isinstance(init, list) else init

    assert ext_embedding.num_embeddings == num_embeddings + n_init
    assert ext_embedding.embedding_dim == embedding_dim
    assert ext_embedding.weight.shape == (num_embeddings + n_init, embedding_dim)
    assert torch.equal(ext_embedding.weight[:num_embeddings], embedding.weight)

    if isinstance(init, list):
        assert torch.equal(ext_embedding.weight[num_embeddings:], embedding.weight[init])

    dummy_indexes = torch.arange(num_embeddings + n_init, dtype=torch.long)
    dummy_prediction = ext_embedding(dummy_indexes)
    dummy_target = torch.randn_like(dummy_prediction)
    dummy_loss = torch.nn.functional.mse_loss(dummy_prediction, dummy_target, reduce=True)
    dummy_loss.backward()
    grad = ext_embedding.weight.grad

    if freeze_prev:
        prev_grad = grad[:num_embeddings]
        new_grad = grad[num_embeddings:]
        assert torch.equal(prev_grad, torch.zeros_like(prev_grad))
        assert not torch.equal(new_grad, torch.zeros_like(new_grad))
    else:
        assert not torch.equal(grad, torch.zeros_like(grad))


def test_embedding_extend_invalid_init(num_embeddings: int = 10, embedding_dim: int = 5):
    embedding = Embedding(num_embeddings, embedding_dim)
    with pytest.raises(ValueError):
        Embedding.extend(embedding, init=0)
    with pytest.raises(TypeError):
        Embedding.extend(embedding, init="invalid")


def test_embedding_save_load(
    save_path: str,
    num_embeddings: int = 10,
    embedding_dim: int = 5,
    init: int = 2,
    freeze_prev: bool = True,
):
    embedding = Embedding(num_embeddings, embedding_dim)
    ext_embedding = Embedding.extend(embedding, init=init, freeze_prev=freeze_prev)

    state_dict_path = join(save_path, "ext_embedding_state_dict.pt")
    torch.save(ext_embedding, state_dict_path)
    embedding = Embedding(num_embeddings, embedding_dim)
    embedding.load_state_dict(torch.load(state_dict_path, weights_only=False).state_dict())



from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
import torch
from torch.distributions import Categorical, Normal
from torch.distributions import kl_divergence as kl
from torch.nn import functional as F

from scvi import REGISTRY_KEYS
from scvi.data import _constants
from scvi.module.base import LossOutput, auto_move_data
from scvi.nn import Decoder, Encoder

from ._classifier import Classifier
from ._utils import broadcast_labels
from ._vae import VAE

if TYPE_CHECKING:
    from collections.abc import Iterable, Sequence
    from typing import Literal

    from torch.distributions import Distribution

    from scvi.model.base import BaseModelClass


class SCANVAE(VAE):
    """Single-cell annotation using variational inference.

    This is an implementation of the scANVI model described in :cite:p:`Xu21`,
    inspired from M1 + M2 model, as described in (https://arxiv.org/pdf/1406.5298.pdf).

    Parameters
    ----------
    n_input
        Number of input genes
    n_batch
        Number of batches
    n_labels
        Number of labels
    n_hidden
        Number of nodes per hidden layer
    n_latent
        Dimensionality of the latent space
    n_layers
        Number of hidden layers used for encoder and decoder NNs
    n_continuous_cov
        Number of continuous covarites
    n_cats_per_cov
        Number of categories for each extra categorical covariate
    dropout_rate
        Dropout rate for neural networks
    dispersion
        One of the following

        * ``'gene'`` - dispersion parameter of NB is constant per gene across cells
        * ``'gene-batch'`` - dispersion can differ between different batches
        * ``'gene-label'`` - dispersion can differ between different labels
        * ``'gene-cell'`` - dispersion can differ for every gene in every cell
    log_variational
        Log(data+1) prior to encoding for numerical stability. Not normalization.
    gene_likelihood
        One of:

        * ``'nb'`` - Negative binomial distribution
        * ``'zinb'`` - Zero-inflated negative binomial distribution
    y_prior
        If None, initialized to uniform probability over cell types
    labels_groups
        Label group designations
    use_labels_groups
        Whether to use the label groups
    linear_classifier
        If `True`, uses a single linear layer for classification instead of a
        multi-layer perceptron.
    classifier_parameters
        Keyword arguments passed into :class:`~scvi.module.Classifier`.
    use_batch_norm
        Whether to use batch norm in layers
    use_layer_norm
        Whether to use layer norm in layers
    linear_classifier
        If ``True``, uses a single linear layer for classification instead of a
        multi-layer perceptron.
    **vae_kwargs
        Keyword args for :class:`~scvi.module.VAE`
    """

    def __init__(
        self,
        n_input: int,
        n_batch: int = 0,
        n_labels: int = 0,
        n_hidden: int = 128,
        n_latent: int = 10,
        n_layers: int = 1,
        n_continuous_cov: int = 0,
        n_cats_per_cov: Iterable[int] | None = None,
        dropout_rate: float = 0.1,
        dispersion: Literal["gene", "gene-batch", "gene-label", "gene-cell"] = "gene",
        log_variational: bool = True,
        gene_likelihood: Literal["zinb", "nb"] = "zinb",
        y_prior: torch.Tensor | None = None,
        labels_groups: Sequence[int] = None,
        use_labels_groups: bool = False,
        linear_classifier: bool = False,
        classifier_parameters: dict | None = None,
        use_batch_norm: Literal["encoder", "decoder", "none", "both"] = "both",
        use_layer_norm: Literal["encoder", "decoder", "none", "both"] = "none",
        **vae_kwargs,
    ):
        super().__init__(
            n_input,
            n_hidden=n_hidden,
            n_latent=n_latent,
            n_layers=n_layers,
            n_continuous_cov=n_continuous_cov,
            n_cats_per_cov=n_cats_per_cov,
            dropout_rate=dropout_rate,
            n_batch=n_batch,
            dispersion=dispersion,
            log_variational=log_variational,
            gene_likelihood=gene_likelihood,
            use_batch_norm=use_batch_norm,
            use_layer_norm=use_layer_norm,
            **vae_kwargs,
        )

        classifier_parameters = classifier_parameters or {}
        use_batch_norm_encoder = use_batch_norm == "encoder" or use_batch_norm == "both"
        use_batch_norm_decoder = use_batch_norm == "decoder" or use_batch_norm == "both"
        use_layer_norm_encoder = use_layer_norm == "encoder" or use_layer_norm == "both"
        use_layer_norm_decoder = use_layer_norm == "decoder" or use_layer_norm == "both"

        self.n_labels = n_labels
        # Classifier takes n_latent as input
        cls_parameters = {
            "n_layers": 0 if linear_classifier else n_layers,
            "n_hidden": 0 if linear_classifier else n_hidden,
            "dropout_rate": dropout_rate,
            "logits": True,
        }
        cls_parameters.update(classifier_parameters)
        self.classifier = Classifier(
            n_latent,
            n_labels=n_labels,
            use_batch_norm=use_batch_norm_encoder,
            use_layer_norm=use_layer_norm_encoder,
            **cls_parameters,
        )

        self.encoder_z2_z1 = Encoder(
            n_latent,
            n_latent,
            n_cat_list=[self.n_labels],
            n_layers=n_layers,
            n_hidden=n_hidden,
            dropout_rate=dropout_rate,
            use_batch_norm=use_batch_norm_encoder,
            use_layer_norm=use_layer_norm_encoder,
            return_dist=True,
        )

        self.decoder_z1_z2 = Decoder(
            n_latent,
            n_latent,
            n_cat_list=[self.n_labels],
            n_layers=n_layers,
            n_hidden=n_hidden,
            use_batch_norm=use_batch_norm_decoder,
            use_layer_norm=use_layer_norm_decoder,
        )

        self.y_prior = torch.nn.Parameter(
            y_prior if y_prior is not None else (1 / n_labels) * torch.ones(1, n_labels),
            requires_grad=False,
        )
        self.use_labels_groups = use_labels_groups
        self.labels_groups = np.array(labels_groups) if labels_groups is not None else None
        if self.use_labels_groups:
            if labels_groups is None:
                raise ValueError("Specify label groups")
            unique_groups = np.unique(self.labels_groups)
            self.n_groups = len(unique_groups)
            if not (unique_groups == np.arange(self.n_groups)).all():
                raise ValueError()
            self.classifier_groups = Classifier(
                n_latent, n_hidden, self.n_groups, n_layers, dropout_rate
            )
            self.groups_index = torch.nn.ParameterList(
                [
                    torch.nn.Parameter(
                        torch.tensor(
                            (self.labels_groups == i).astype(np.uint8),
                            dtype=torch.uint8,
                        ),
                        requires_grad=False,
                    )
                    for i in range(self.n_groups)
                ]
            )

    @auto_move_data
    def classify(
        self,
        x: torch.Tensor,
        batch_index: torch.Tensor | None = None,
        cont_covs: torch.Tensor | None = None,
        cat_covs: torch.Tensor | None = None,
        use_posterior_mean: bool = True,
    ) -> torch.Tensor:
        """Forward pass through the encoder and classifier.

        Parameters
        ----------
        x
            Tensor of shape ``(n_obs, n_vars)``.
        batch_index
            Tensor of shape ``(n_obs,)`` denoting batch indices.
        cont_covs
            Tensor of shape ``(n_obs, n_continuous_covariates)``.
        cat_covs
            Tensor of shape ``(n_obs, n_categorical_covariates)``.
        use_posterior_mean
            Whether to use the posterior mean of the latent distribution for
            classification.

        Returns
        -------
        Tensor of shape ``(n_obs, n_labels)`` denoting logit scores per label.
        Before v1.1, this method by default returned probabilities per label,
        see #2301 for more details.
        """
        if self.log_variational:
            x = torch.log1p(x)

        if cont_covs is not None and self.encode_covariates:
            encoder_input = torch.cat((x, cont_covs), dim=-1)
        else:
            encoder_input = x
        if cat_covs is not None and self.encode_covariates:
            categorical_input = torch.split(cat_covs, 1, dim=1)
        else:
            categorical_input = ()

        qz, z = self.z_encoder(encoder_input, batch_index, *categorical_input)
        z = qz.loc if use_posterior_mean else z

        if self.use_labels_groups:
            w_g = self.classifier_groups(z)
            unw_y = self.classifier(z)
            w_y = torch.zeros_like(unw_y)
            for i, group_index in enumerate(self.groups_index):
                unw_y_g = unw_y[:, group_index]
                w_y[:, group_index] = unw_y_g / (unw_y_g.sum(dim=-1, keepdim=True) + 1e-8)
                w_y[:, group_index] *= w_g[:, [i]]
        else:
            w_y = self.classifier(z)
        return w_y

    @auto_move_data
    def classification_loss(
        self, labelled_dataset: dict[str, torch.Tensor]
    ) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        x = labelled_dataset[REGISTRY_KEYS.X_KEY]  # (n_obs, n_vars)
        y = labelled_dataset[REGISTRY_KEYS.LABELS_KEY]  # (n_obs, 1)
        batch_idx = labelled_dataset[REGISTRY_KEYS.BATCH_KEY]
        cont_key = REGISTRY_KEYS.CONT_COVS_KEY
        cont_covs = labelled_dataset[cont_key] if cont_key in labelled_dataset.keys() else None

        cat_key = REGISTRY_KEYS.CAT_COVS_KEY
        cat_covs = labelled_dataset[cat_key] if cat_key in labelled_dataset.keys() else None
        # NOTE: prior to v1.1, this method returned probabilities per label by
        # default, see #2301 for more details
        logits = self.classify(
            x, batch_index=batch_idx, cat_covs=cat_covs, cont_covs=cont_covs
        )  # (n_obs, n_labels)
        ce_loss = F.cross_entropy(
            logits,
            y.view(-1).long(),
        )
        return ce_loss, y, logits

    def loss(
        self,
        tensors: dict[str, torch.Tensor],
        inference_outputs: dict[str, torch.Tensor | Distribution | None],
        generative_ouputs: dict[str, Distribution | None],
        kl_weight: float = 1.0,
        labelled_tensors: dict[str, torch.Tensor] | None = None,
        classification_ratio: float | None = None,
    ):
        """Compute the loss."""
        px: Distribution = generative_ouputs["px"]
        qz1: torch.Tensor = inference_outputs["qz"]
        z1: torch.Tensor = inference_outputs["z"]
        x: torch.Tensor = tensors[REGISTRY_KEYS.X_KEY]
        batch_index: torch.Tensor = tensors[REGISTRY_KEYS.BATCH_KEY]

        ys, z1s = broadcast_labels(z1, n_broadcast=self.n_labels)
        qz2, z2 = self.encoder_z2_z1(z1s, ys)
        pz1_m, pz1_v = self.decoder_z1_z2(z2, ys)
        reconst_loss = -px.log_prob(x).sum(-1)

        # KL Divergence
        mean = torch.zeros_like(qz2.loc)
        scale = torch.ones_like(qz2.scale)

        kl_divergence_z2 = kl(qz2, Normal(mean, scale)).sum(dim=-1)
        loss_z1_unweight = -Normal(pz1_m, torch.sqrt(pz1_v)).log_prob(z1s).sum(dim=-1)
        loss_z1_weight = qz1.log_prob(z1).sum(dim=-1)

        probs = self.classifier(z1)
        if self.classifier.logits:
            probs = F.softmax(probs, dim=-1)

        if z1.ndim == 2:
            loss_z1_unweight_ = loss_z1_unweight.view(self.n_labels, -1).t()
            kl_divergence_z2_ = kl_divergence_z2.view(self.n_labels, -1).t()
        else:
            loss_z1_unweight_ = torch.transpose(
                loss_z1_unweight.view(z1.shape[0], self.n_labels, -1), -1, -2
            )
            kl_divergence_z2_ = torch.transpose(
                kl_divergence_z2.view(z1.shape[0], self.n_labels, -1), -1, -2
            )
        reconst_loss += loss_z1_weight + (loss_z1_unweight_ * probs).sum(dim=-1)
        kl_divergence = (kl_divergence_z2_ * probs).sum(dim=-1)
        kl_divergence += kl(
            Categorical(probs=probs),
            Categorical(
                probs=self.y_prior.repeat(probs.size(0), probs.size(1), 1)
                if len(probs.size()) == 3
                else self.y_prior.repeat(probs.size(0), 1)
            ),
        )

        if not self.use_observed_lib_size:
            ql = inference_outputs["ql"]
            (
                local_library_log_means,
                local_library_log_vars,
            ) = self._compute_local_library_params(batch_index)

            kl_divergence_l = kl(
                ql,
                Normal(local_library_log_means, torch.sqrt(local_library_log_vars)),
            ).sum(dim=1)
        else:
            kl_divergence_l = torch.zeros_like(kl_divergence)

        kl_divergence += kl_divergence_l

        loss = torch.mean(reconst_loss + kl_divergence * kl_weight)

        if labelled_tensors is not None:
            ce_loss, true_labels, logits = self.classification_loss(labelled_tensors)

            loss += ce_loss * classification_ratio
            return LossOutput(
                loss=loss,
                reconstruction_loss=reconst_loss,
                kl_local=kl_divergence,
                classification_loss=ce_loss,
                true_labels=true_labels,
                logits=logits,
            )
        return LossOutput(loss=loss, reconstruction_loss=reconst_loss, kl_local=kl_divergence)

    def on_load(self, model: BaseModelClass):
        manager = model.get_anndata_manager(model.adata, required=True)
        source_version = manager._source_registry[_constants._SCVI_VERSION_KEY]
        version_split = source_version.split(".")

        if int(version_split[0]) >= 1 and int(version_split[1]) >= 1:
            return

        # need this if <1.1 model is resaved with >=1.1 as new registry is
        # updated on setup
        manager.registry[_constants._SCVI_VERSION_KEY] = source_version

        # pre 1.1 logits fix
        model_kwargs = model.init_params_.get("model_kwargs", {})
        cls_params = model_kwargs.get("classifier_parameters", {})
        user_logits = cls_params.get("logits", False)

        if not user_logits:
            self.classifier.logits = False
            self.classifier.classifier.append(torch.nn.Softmax(dim=-1))



from __future__ import annotations

from os.path import join
from typing import TYPE_CHECKING

from anndata import AnnData
from lightning.pytorch.callbacks import Callback
from lightning.pytorch.loggers import TensorBoardLogger
from mudata import MuData
from ray.tune import Tuner

if TYPE_CHECKING:
    from typing import Any, Literal

    from lightning.pytorch import LightningDataModule
    from ray.tune import ResultGrid
    from ray.tune.schedulers import TrialScheduler
    from ray.tune.search import SearchAlgorithm

    from scvi._types import AnnOrMuData
    from scvi.model.base import BaseModelClass

_ASHA_DEFAULT_KWARGS = {
    "max_t": 100,
    "grace_period": 1,
    "reduction_factor": 2,
}


class AutotuneExperiment:
    """``BETA`` Track hyperparameter tuning experiments.

    Parameters
    ----------
    model_cls
        Model class on which to tune hyperparameters. Must implement a constructor and a ``train``
        method.
    data
        :class:`~anndata.AnnData` or :class:`~mudata.MuData` that has been setup with
        ``model_cls`` or a :class:`~lightning.pytorch.core.LightningDataModule` (``EXPERIMENTAL``).
    metrics
        Either a single metric or a list of metrics to track during the experiment. If a list is
        provided, the primary metric will be the first element in the list.
    mode
        Optimization mode for the primary metric. One of ``"min"`` or ``"max"``.
    search_space
        Dictionary of hyperparameter names and their respective search spaces. See the
        `API <https://docs.ray.io/en/latest/tune/api/search_space.html>`_ for available search
        specifications. Must only contain the following top-level keys:

        - ``"model_params"``: parameters to pass to the model constructor.
        - ``"train_params"``: parameters to pass to the model's ``train`` method.

        Passed into :class:`~ray.tune.Tuner` as ``param_space``.
    num_samples
        Total number of hyperparameter configurations to sample from the search space. Passed into
        :class:`~ray.tune.tune_config.TuneConfig`.
    scheduler
        Ray Tune scheduler to use. One of the following:

        - ``"asha"``: :class:`~ray.tune.schedulers.AsyncHyperBandScheduler`
        - ``"hyperband"``: :class:`~ray.tune.schedulers.HyperBandScheduler`
        - ``"median"``: :class:`~ray.tune.schedulers.MedianStoppingRule`
        - ``"fifo"``: :class:`~ray.tune.schedulers.FIFOScheduler`

        Configured with reasonable defaults, which can be overridden with ``scheduler_kwargs``.
    searcher
        Ray Tune search algorithm to use. One of the following:

        - ``"hyperopt"``: :class:`~ray.tune.search.hyperopt.HyperOptSearch`
        - ``"random"``: :class:`~ray.tune.search.basic_variant.BasicVariantGenerator`

        Configured with reasonable defaults, which can be overridden with ``searcher_kwargs``.
    seed
        Random seed to use for the experiment. Propagated to :attr:`~scvi.settings.seed` and search
        algorithms. If not provided, defaults to :attr:`~scvi.settings.seed`.
    resources
        Dictionary of resources to allocate per trial in the experiment. Available keys include:

        - ``"cpu"``: number of CPU cores
        - ``"gpu"``: number of GPUs
        - ``"memory"``: amount of memory

        Passed into :func:`~ray.tune.with_resources`.
    name
        Name of the experiment, used for logging purposes. Defaults to a unique ID.
    logging_dir
        Base directory to store experiment logs. Defaults to :attr:`~scvi.settings.logging_dir`.
    scheduler_kwargs
        Additional keyword arguments to pass to the scheduler.
    searcher_kwargs
        Additional keyword arguments to pass to the search algorithm.

    Notes
    -----
    Lifecycle: beta

    See Also
    --------
    :func:`~scvi.autotune.run_autotune`
    """

    def __init__(
        self,
        model_cls: BaseModelClass,
        data: AnnOrMuData | LightningDataModule,
        metrics: str | list[str],
        mode: Literal["min", "max"],
        search_space: dict[str, dict[Literal["model_params", "train_params"], dict[str, Any]]],
        num_samples: int,
        scheduler: Literal["asha", "hyperband", "median", "fifo"] = "asha",
        searcher: Literal["hyperopt", "random"] = "hyperopt",
        seed: int | None = None,
        resources: dict[Literal["cpu", "gpu", "memory"], float] | None = None,
        name: str | None = None,
        logging_dir: str | None = None,
        scheduler_kwargs: dict | None = None,
        searcher_kwargs: dict | None = None,
    ) -> None:
        self.model_cls = model_cls
        self.data = data
        self.metrics = metrics
        self.mode = mode
        self.search_space = search_space
        self.num_samples = num_samples
        self.seed = seed
        self.scheduler_kwargs = scheduler_kwargs
        self.searcher_kwargs = searcher_kwargs
        self.scheduler = scheduler
        self.searcher = searcher
        self.resources = resources
        self.name = name
        self.logging_dir = logging_dir

    @property
    def id(self) -> str:
        """Unique identifier for the experiment."""
        from uuid import uuid4

        if not hasattr(self, "_id"):
            self._id = str(uuid4())
        return self._id

    @property
    def model_cls(self) -> BaseModelClass:
        """Model class on which to tune hyperparameters."""
        return self._model_cls

    @model_cls.setter
    def model_cls(self, value: BaseModelClass) -> None:
        if hasattr(self, "_model_cls"):
            raise AttributeError("Cannot reassign `model_cls`")
        self._model_cls = value

    @property
    def data(self) -> AnnOrMuData | LightningDataModule:
        """Data on which to tune hyperparameters."""
        return self._data

    @data.setter
    def data(self, value: AnnOrMuData | LightningDataModule) -> None:
        from scvi.data._constants import _SETUP_ARGS_KEY, _SETUP_METHOD_NAME

        if hasattr(self, "_data"):
            raise AttributeError("Cannot reassign `data`")

        self._data = value
        if isinstance(value, AnnData | MuData):
            data_manager = self.model_cls._get_most_recent_anndata_manager(value, required=True)
            self._setup_method_name = data_manager._registry.get(
                _SETUP_METHOD_NAME, "setup_anndata"
            )
            self._setup_method_args = data_manager._get_setup_method_args().get(
                _SETUP_ARGS_KEY, {}
            )

    @property
    def setup_method_name(self) -> str:
        """Either ``"setup_anndata"`` or ``"setup_mudata"``."""
        if not hasattr(self, "_setup_method_name"):
            raise AttributeError("`setup_method_name` not available.")
        return self._setup_method_name

    @property
    def setup_method_args(self) -> dict[str, Any]:
        """Keyword arguments for the setup method."""
        if not hasattr(self, "_setup_method_args"):
            raise AttributeError("`setup_method_args` not available.")
        return self._setup_method_args

    @property
    def metrics(self) -> list[str]:
        """Metrics to track during the experiment."""
        if not hasattr(self, "_metrics"):
            raise AttributeError("`metrics` not yet available.")
        return self._metrics

    @metrics.setter
    def metrics(self, value: str | list[str]) -> None:
        if hasattr(self, "_metrics"):
            raise AttributeError("Cannot reassign `metrics`")
        elif value is None:
            raise ValueError("`metrics` must not be `None`")
        elif not isinstance(value, str) and not isinstance(value, list):
            raise TypeError("`metrics` must be a string or a list of strings")
        elif isinstance(value, str):
            value = [value]
        elif isinstance(value, list) and len(value) < 1:
            raise ValueError("`metrics` must not be empty")
        self._metrics = value

    @property
    def mode(self) -> Literal["min", "max"]:
        """Optimization mode for the primary metric."""
        if not hasattr(self, "_mode"):
            raise AttributeError("`mode` not yet available.")
        return self._mode

    @mode.setter
    def mode(self, value: Literal["min", "max"]) -> None:
        if hasattr(self, "_mode"):
            raise AttributeError("Cannot reassign `mode`")
        elif value not in ["min", "max"]:
            raise ValueError("`mode` must be either 'min' or 'max'")
        self._mode = value

    @property
    def search_space(
        self,
    ) -> dict[str, dict[Literal["model_params", "train_params"], dict[str, Any]]]:
        """Search space for hyperparameters."""
        if not hasattr(self, "_search_space"):
            raise AttributeError("`search_space` not yet available.")
        return self._search_space

    @search_space.setter
    def search_space(self, value: dict[str, Any]) -> None:
        if hasattr(self, "_search_space"):
            raise AttributeError("Cannot reassign `search_space`")
        elif not isinstance(value, dict):
            raise TypeError("`search_space` must be a dictionary")
        elif len(value) == 0:
            raise ValueError("`search_space` must not be empty")
        elif any(key not in ["model_params", "train_params"] for key in value.keys()):
            raise KeyError(
                "`search_space` must only contain 'model_params' and 'train_params' keys"
            )

        self._search_space = value

    @property
    def num_samples(self) -> int:
        """Total number of hyperparameter configurations to sample."""
        if not hasattr(self, "_num_samples"):
            raise AttributeError("`num_samples` not yet available.")
        return self._num_samples

    @num_samples.setter
    def num_samples(self, value: int | None) -> None:
        if hasattr(self, "_num_samples"):
            raise AttributeError("Cannot reassign `num_samples`")
        elif not isinstance(value, int):
            raise TypeError("`num_samples` must be an integer")
        self._num_samples = value

    @property
    def scheduler(self) -> TrialScheduler:
        """Ray Tune scheduler to use."""
        if not hasattr(self, "_scheduler"):
            raise AttributeError("`scheduler` not yet available.")
        return self._scheduler

    @scheduler.setter
    def scheduler(self, value: Literal["asha", "hyperband", "median", "fifo"]) -> None:
        from ray.tune.schedulers import (
            AsyncHyperBandScheduler,
            FIFOScheduler,
            HyperBandScheduler,
            MedianStoppingRule,
        )

        if hasattr(self, "_scheduler"):
            raise AttributeError("Cannot reassign `scheduler`")
        elif not isinstance(value, str):
            raise TypeError("`scheduler` must be a string")
        elif value not in ["asha", "hyperband", "median", "fifo"]:
            raise ValueError("`scheduler` must be one of 'asha', 'hyperband', 'median', 'fifo'")

        kwargs = {
            "metric": self.metrics[0],
            "mode": self.mode,
        }
        if value == "asha":
            kwargs.update(_ASHA_DEFAULT_KWARGS)
            scheduler_cls = AsyncHyperBandScheduler
        elif value == "hyperband":
            scheduler_cls = HyperBandScheduler
        elif value == "median":
            scheduler_cls = MedianStoppingRule
        else:
            kwargs = {}
            scheduler_cls = FIFOScheduler

        kwargs.update(self.scheduler_kwargs)
        self._scheduler = scheduler_cls(**kwargs)

    @property
    def scheduler_kwargs(self) -> dict[str, Any]:
        """Additional keyword arguments to pass to the scheduler."""
        if not hasattr(self, "_scheduler_kwargs"):
            raise AttributeError("`scheduler_kwargs` not yet available.")
        return self._scheduler_kwargs

    @scheduler_kwargs.setter
    def scheduler_kwargs(self, value: dict | None) -> None:
        if hasattr(self, "_scheduler_kwargs"):
            raise AttributeError("Cannot reassign `scheduler_kwargs`")
        elif value is not None and not isinstance(value, dict):
            raise TypeError("`scheduler_kwargs` must be a dictionary")
        self._scheduler_kwargs = value or {}

    @property
    def searcher(self) -> SearchAlgorithm:
        """Ray Tune search algorithm to use."""
        if not hasattr(self, "_searcher"):
            raise AttributeError("`searcher` not yet available.")
        return self._searcher

    @searcher.setter
    def searcher(self, value: Literal["hyperopt", "random"]) -> None:
        from ray.tune.search import BasicVariantGenerator
        from ray.tune.search.hyperopt import HyperOptSearch

        if hasattr(self, "_searcher"):
            raise AttributeError("Cannot reassign `searcher`")
        elif not isinstance(value, str):
            raise TypeError("`searcher` must be a string")
        elif value not in ["hyperopt", "random"]:
            raise ValueError("`searcher` must be one of 'hyperopt', 'random'")

        if value == "random":
            kwargs = {"random_state": self.seed}
            searcher_cls = BasicVariantGenerator
        else:
            kwargs = {
                "metric": self.metrics[0],
                "mode": self.mode,
                "random_state_seed": self.seed,
            }
            searcher_cls = HyperOptSearch

        kwargs.update(self.searcher_kwargs)
        self._searcher = searcher_cls(**kwargs)

    @property
    def searcher_kwargs(self) -> dict[str, Any]:
        """Additional keyword arguments to pass to the search algorithm."""
        if not hasattr(self, "_searcher_kwargs"):
            raise AttributeError("`searcher_kwargs` not yet available.")
        return self._searcher_kwargs

    @searcher_kwargs.setter
    def searcher_kwargs(self, value: dict | None) -> None:
        if hasattr(self, "_searcher_kwargs"):
            raise AttributeError("Cannot reassign `searcher_kwargs`")
        elif value is not None and not isinstance(value, dict):
            raise TypeError("`searcher_kwargs` must be a dictionary")
        self._searcher_kwargs = value or {}

    @property
    def seed(self) -> int | None:
        """Random seed to use for the experiment."""
        if not hasattr(self, "_seed"):
            raise AttributeError("`seed` not yet available.")
        return self._seed

    @seed.setter
    def seed(self, value: int | None) -> None:
        from scvi import settings

        if hasattr(self, "_seed"):
            raise AttributeError("Cannot reassign `seed`")
        elif value is not None and not isinstance(value, int):
            raise TypeError("`seed` must be an integer")
        self._seed = value or settings.seed

    @property
    def resources(self) -> dict[Literal["cpu", "gpu", "memory"], float] | None:
        """Resources to allocate per trial in the experiment."""
        if not hasattr(self, "_resources"):
            raise AttributeError("`resources` not yet available.")
        return self._resources

    @resources.setter
    def resources(self, value: dict[Literal["cpu", "gpu", "memory"], float] | None) -> None:
        if hasattr(self, "_resources"):
            raise AttributeError("Cannot reassign `resources`")
        elif value is not None and not isinstance(value, dict):
            raise TypeError("`resources` must be a dictionary or `None`")
        self._resources = value or {}

    @property
    def name(self) -> str:
        """Name of the experiment."""
        if not hasattr(self, "_name"):
            raise AttributeError("`name` not yet available.")
        return self._name

    @name.setter
    def name(self, value: str | None) -> None:
        if hasattr(self, "_name"):
            raise AttributeError("Cannot reassign `name`")
        elif value is not None and not isinstance(value, str):
            raise TypeError("`name` must be a string or `None`")

        if value is None:
            default = f"{self._model_cls.__name__.lower()}_"
            default += self.id
        self._name = value or default

    @property
    def logging_dir(self) -> str:
        """Base directory to store experiment logs."""
        if not hasattr(self, "_logging_dir"):
            raise AttributeError("`logging_dir` not yet available.")
        return self._logging_dir

    @logging_dir.setter
    def logging_dir(self, value: str | None) -> None:
        from scvi import settings

        if hasattr(self, "_logging_dir"):
            raise AttributeError("Cannot reassign `logging_dir`")
        elif value is not None and not isinstance(value, str):
            raise TypeError("`logging_dir` must be a string")
        self._logging_dir = value or join(settings.logging_dir, self.name)

    @property
    def metrics_callback(self) -> Callback:
        from ray.tune.integration.pytorch_lightning import TuneReportCheckpointCallback

        if not hasattr(self, "_metrics"):
            raise AttributeError("`metrics_callback` not yet available.")

        # this is to get around lightning import changes
        callback_cls = type(
            "_TuneReportCheckpointCallback",
            (TuneReportCheckpointCallback, Callback),
            {},
        )
        return callback_cls(metrics=self.metrics, on="validation_end", save_checkpoints=False)

    @property
    def result_grid(self) -> ResultGrid:
        """Result grid for the experiment."""
        if not hasattr(self, "_result_grid"):
            raise AttributeError("`result_grid` not yet available.")
        return self._result_grid

    @result_grid.setter
    def result_grid(self, value: ResultGrid) -> None:
        if hasattr(self, "_result_grid"):
            raise AttributeError("Cannot reassign `result_grid`")
        self._result_grid = value

    def __repr__(self) -> str:
        return f"Experiment {self.name}"

    def get_tuner(self) -> Tuner:
        """Configure a :class:`~ray.tune.Tuner` from this experiment."""
        from ray.train import RunConfig
        from ray.tune import with_parameters, with_resources
        from ray.tune.tune_config import TuneConfig

        trainable = with_parameters(_trainable, experiment=self)
        trainable = with_resources(trainable, resources=self.resources)

        tune_config = TuneConfig(
            scheduler=self.scheduler,
            search_alg=self.searcher,
            num_samples=self.num_samples,
        )
        run_config = RunConfig(
            name=self.name,
            storage_path=self.logging_dir,
            log_to_file=True,
            verbose=1,
        )
        return Tuner(
            trainable=trainable,
            param_space=self.search_space,
            tune_config=tune_config,
            run_config=run_config,
        )

    def get_logger(self, trial_name: str) -> TensorBoardLogger:
        """Configure TensorBoard logger for a trial in this experiment."""
        return TensorBoardLogger(join(self.logging_dir, f"{trial_name}_tensorboard"))


def _trainable(
    param_sample: dict[str, dict[Literal["model_params", "train_params"], dict[str, Any]]],
    experiment: AutotuneExperiment,
) -> None:
    """Implements a Ray Tune trainable function for an :class:`~scvi.autotune.AutotuneExperiment`.

    Setup on the :class:`~anndata.AnnData` or :class:`~mudata.MuData` has to be performed since Ray
    opens a new process per trial and thus the initial setup on the main process is not
    transferred.

    Parameters
    ----------
    param_sample
        Hyperparameter configuration on which to train. Note: this is different from
        :attr:`~scvi.autotune.AutotuneExperiment.search_space` in that it is a single configuration
        sampled from the search space, not the specification of the search space itself.
    experiment
        :class:`~scvi.autotune.AutotuneExperiment` to evaluate.

    Notes
    -----
    See the Ray Tune
    `documentation <https://docs.ray.io/en/latest/tune/api/trainable.html#function-trainable-api>`_
    for more details.
    """
    from ray.train import get_context

    from scvi import settings

    model_params, train_params = (
        param_sample.get("model_params", {}),
        param_sample.get("train_params", {}),
    )
    train_params = {
        "accelerator": "auto",
        "devices": "auto",
        "check_val_every_n_epoch": 1,
        "enable_progress_bar": False,
        "logger": experiment.get_logger(get_context().get_trial_name()),
        "callbacks": [experiment.metrics_callback],
        **train_params,
    }

    settings.seed = experiment.seed
    if isinstance(experiment.data, AnnData | MuData):
        getattr(experiment.model_cls, experiment.setup_method_name)(
            experiment.data,
            **experiment.setup_method_args,
        )
        model = experiment.model_cls(experiment.data, **model_params)
        model.train(**train_params)
    else:
        model = experiment.model_cls(**train_params)
        model.train(datamodule=experiment.data, **train_params)



from math import ceil, floor

import numpy as np
import pytest
from tests.data.utils import generic_setup_adata_manager

import scvi
from scvi.dataloaders import BatchDistributedSampler


def test_batchdistributedsampler_init(
    batch_size: int = 128,
    n_batches: int = 2,
):
    adata = scvi.data.synthetic_iid(batch_size=batch_size, n_batches=n_batches)
    manager = generic_setup_adata_manager(adata)
    dataset = manager.create_torch_dataset()

    sampler = BatchDistributedSampler(
        dataset,
        num_replicas=1,
        rank=0,
        batch_size=batch_size,
        shuffle=True,
        drop_last=True,
        drop_dataset_tail=True,
    )
    assert sampler.batch_size == batch_size
    assert sampler.rank == 0
    assert sampler.num_replicas == 1
    assert sampler.shuffle
    assert sampler.drop_last_batch  # drop_last
    assert sampler.drop_last  # drop_dataset_tail


@pytest.mark.parametrize("drop_last", [True, False])
@pytest.mark.parametrize("drop_dataset_tail", [True, False])
def test_batchdistributedsampler_drop_last(
    drop_last: bool,
    drop_dataset_tail: bool,
    batch_size: int = 128,
    n_batches: int = 3,
    num_replicas: int = 2,
):
    """Expected behavior:

    drop_last: False
    drop_dataset_tail: False
    - samplers see m = ceil(n_obs / num_replicas) observations
    - samplers have b = ceil(m / batch_size) batches

    drop_last: True
    drop_dataset_tail: False
    - samplers see m = ceil(n_obs / num_replicas) observations
    - samplers have b = floor(m / batch_size) batches
    - error if b == 0

    drop_last: False
    drop_dataset_tail: True
    - samplers see m = floor(n_obs / num_replicas) observations
    - samplers have b = ceil(m / batch_size) batches
    - error if m == 0

    drop_last: True
    drop_dataset_tail: True
    - samplers see m = floor(n_obs / num_replicas) observations
    - samplers have b = floor(m / batch_size) batches
    - error if m == 0 or b == 0
    """
    adata = scvi.data.synthetic_iid(batch_size=batch_size, n_batches=n_batches)
    manager = generic_setup_adata_manager(adata)
    dataset = manager.create_torch_dataset()

    def check_samplers(samplers: list, sampler_batch_size: int):
        sampler_indices = [list(sampler) for sampler in samplers]

        if drop_dataset_tail:
            n_obs_per_sampler = floor(len(dataset) / num_replicas)
        else:
            n_obs_per_sampler = ceil(len(dataset) / num_replicas)

        if drop_last:
            n_batches_per_sampler = floor(n_obs_per_sampler / sampler_batch_size)
            batch_sizes = [sampler_batch_size] * n_batches_per_sampler
        else:
            n_batches_per_sampler = ceil(n_obs_per_sampler / sampler_batch_size)
            if n_obs_per_sampler % sampler_batch_size == 0:
                batch_sizes = [sampler_batch_size] * n_batches_per_sampler
            else:
                batch_sizes = [sampler_batch_size] * (n_batches_per_sampler - 1) + [
                    n_obs_per_sampler % sampler_batch_size
                ]

        effective_n_obs_per_sampler = sum(batch_sizes)

        assert len(sampler_indices) == num_replicas
        for batch_indices in sampler_indices:
            all_indices = np.concatenate(batch_indices)

            assert len(batch_indices) == n_batches_per_sampler
            assert len(all_indices) == effective_n_obs_per_sampler
            assert [len(indices) for indices in batch_indices] == batch_sizes

    for sampler_batch_size in [batch_size, batch_size - 1, batch_size + 1]:
        samplers = [
            BatchDistributedSampler(
                dataset,
                num_replicas=num_replicas,
                rank=i,
                batch_size=sampler_batch_size,
                drop_last=drop_last,
                drop_dataset_tail=drop_dataset_tail,
            )
            for i in range(num_replicas)
        ]
        check_samplers(samplers, sampler_batch_size)


def test_batchdistributedsampler_indices(
    batch_size: int = 128,
    n_batches: int = 3,
    num_replicas: int = 2,
):
    adata = scvi.data.synthetic_iid(batch_size=batch_size, n_batches=n_batches)
    manager = generic_setup_adata_manager(adata)
    dataset = manager.create_torch_dataset()

    samplers = [
        BatchDistributedSampler(
            dataset,
            num_replicas=num_replicas,
            rank=i,
            batch_size=batch_size,
        )
        for i in range(num_replicas)
    ]
    sampler_indices = [list(sampler) for sampler in samplers]
    sampler_indices = [set(np.concatenate(indices)) for indices in sampler_indices]

    # check that indices are non-overlapping
    for i in range(num_replicas):
        for j in range(i + 1, num_replicas):
            assert len(sampler_indices[i].intersection(sampler_indices[j])) == 0

    # check that all indices are covered
    covered_indices = np.concatenate([np.array(list(indices)) for indices in sampler_indices])
    assert len(covered_indices) == len(dataset)



import pytest
import torch

from scvi.distributions import BetaBinomial


def test_beta_binomial_distribution():
    alpha = torch.ones(size=(2,))
    beta = torch.ones_like(alpha)

    num_trials = 100
    total_count = torch.ones_like(alpha) * num_trials
    x = torch.randint_like(total_count, high=num_trials)

    dist_param_1 = BetaBinomial(
        total_count=total_count, alpha=alpha, beta=beta, validate_args=True
    )

    log_p_alpha_beta = dist_param_1.log_prob(x)

    mu = alpha / (alpha + beta)
    gamma = 1 / (alpha + beta + 1)

    dist_param_2 = BetaBinomial(total_count=total_count, mu=mu, gamma=gamma, validate_args=True)

    log_p_mu_gamma = dist_param_2.log_prob(x)
    assert (log_p_alpha_beta - log_p_mu_gamma).abs().max().item() <= 1e-8

    # Should fail with value outside of distribution's support
    with pytest.raises(ValueError):
        dist_param_1.log_prob(-x)

    # Should fail as no parameterization is specified
    with pytest.raises(ValueError):
        BetaBinomial(
            total_count=total_count,
        )

    # Should fail as two full parameterizations are provided
    with pytest.raises(ValueError):
        BetaBinomial(total_count=total_count, alpha=alpha, beta=beta, mu=mu, gamma=gamma)

    # Should fail without a complete parameterization 1 or 2
    with pytest.raises(ValueError):
        BetaBinomial(total_count=total_count, alpha=alpha, gamma=gamma)



import os
import tarfile

import pytest
import scanpy as sc

from scvi.data import dataset_10x, organize_cite_seq_10x

from .utils import unsupervised_training_one_epoch


def test_populate_and_train_one_v1(save_path):
    sp = os.path.join(save_path, "10X")
    dataset = dataset_10x(
        dataset_name="cd4_t_helper",
        remove_extracted_data=True,
        save_path=sp,
    )
    unsupervised_training_one_epoch(dataset)


def test_brain_small(save_path):
    sp = os.path.join(save_path, "10X")
    dataset = dataset_10x(
        dataset_name="neuron_9k",
        save_path=sp,
        remove_extracted_data=True,
    )
    unsupervised_training_one_epoch(dataset)


def test_pbmc_cite(save_path):
    file_path = os.path.join(
        save_path, "10X/pbmc_10k_protein_v3/filtered_feature_bc_matrix.tar.gz"
    )
    sp = os.path.join(save_path, "10X/pbmc_10k_protein_v3/")
    tar = tarfile.open(file_path, "r:gz")
    tar.extractall(path=sp)
    tar.close()
    dataset = sc.read_10x_mtx(os.path.join(sp, "filtered_feature_bc_matrix"), gex_only=False)
    organize_cite_seq_10x(dataset)
    unsupervised_training_one_epoch(dataset)


@pytest.mark.internet
def test_download_dataset_10x(save_path):
    dataset_10x("hgmm_1k_v3", save_path=save_path)



import os
import pickle

import numpy as np
import pytest
import torch

from scvi.data import synthetic_iid
from scvi.model import PEAKVI
from scvi.utils import attrdict


def test_saving_and_loading(save_path):
    def legacy_save(
        model,
        dir_path,
        prefix=None,
        overwrite=False,
        save_anndata=False,
        **anndata_write_kwargs,
    ):
        if not os.path.exists(dir_path) or overwrite:
            os.makedirs(dir_path, exist_ok=overwrite)
        else:
            raise ValueError(
                f"{dir_path} already exists. Please provide an unexisting directory for saving."
            )

        file_name_prefix = prefix or ""

        if save_anndata:
            model.adata.write(
                os.path.join(dir_path, f"{file_name_prefix}adata.h5ad"),
                **anndata_write_kwargs,
            )

        model_save_path = os.path.join(dir_path, f"{file_name_prefix}model_params.pt")
        attr_save_path = os.path.join(dir_path, f"{file_name_prefix}attr.pkl")
        varnames_save_path = os.path.join(dir_path, f"{file_name_prefix}var_names.csv")

        torch.save(model.module.state_dict(), model_save_path)

        var_names = model.adata.var_names.astype(str)
        var_names = var_names.to_numpy()
        np.savetxt(varnames_save_path, var_names, fmt="%s")

        # get all the user attributes
        user_attributes = model._get_user_attributes()
        # only save the public attributes with _ at the very end
        user_attributes = {a[0]: a[1] for a in user_attributes if a[0][-1] == "_"}

        with open(attr_save_path, "wb") as f:
            pickle.dump(user_attributes, f)

    def test_save_load_model(cls, adata, save_path, prefix=None):
        cls.setup_anndata(adata, batch_key="batch", labels_key="labels")
        model = cls(adata, latent_distribution="normal")
        model.train(1, train_size=0.2)
        z1 = model.get_latent_representation(adata)
        test_idx1 = model.validation_indices
        model.save(save_path, overwrite=True, save_anndata=True, prefix=prefix)
        model.view_setup_args(save_path, prefix=prefix)
        model = cls.load(save_path, prefix=prefix)
        model.get_latent_representation()

        # Load with mismatched genes.
        tmp_adata = synthetic_iid(
            n_genes=200,
        )
        with pytest.raises(ValueError):
            cls.load(save_path, adata=tmp_adata, prefix=prefix)

        # Load with different batches.
        tmp_adata = synthetic_iid()
        tmp_adata.obs["batch"] = tmp_adata.obs["batch"].cat.rename_categories(
            ["batch_2", "batch_3"]
        )
        with pytest.raises(ValueError):
            cls.load(save_path, adata=tmp_adata, prefix=prefix)

        model = cls.load(save_path, adata=adata, prefix=prefix)
        assert "batch" in model.adata_manager.data_registry
        assert model.adata_manager.data_registry.batch == attrdict(
            {"attr_name": "obs", "attr_key": "_scvi_batch"}
        )

        z2 = model.get_latent_representation()
        test_idx2 = model.validation_indices
        np.testing.assert_array_equal(z1, z2)
        np.testing.assert_array_equal(test_idx1, test_idx2)
        assert model.is_trained is True

        # Test legacy loading
        legacy_save_path = os.path.join(save_path, "legacy/")
        legacy_save(model, legacy_save_path, overwrite=True, save_anndata=True, prefix=prefix)
        with pytest.raises(ValueError):
            cls.load(legacy_save_path, adata=adata, prefix=prefix)
        cls.convert_legacy_save(
            legacy_save_path,
            legacy_save_path,
            overwrite=True,
            prefix=prefix,
        )
        m = cls.load(legacy_save_path, adata=adata, prefix=prefix)
        m.train(1)

    save_path = os.path.join(save_path, "tmp")
    adata = synthetic_iid()

    test_save_load_model(PEAKVI, adata, save_path, prefix=f"{PEAKVI.__name__}_")


def test_peakvi():
    data = synthetic_iid()
    PEAKVI.setup_anndata(
        data,
        batch_key="batch",
    )
    vae = PEAKVI(
        data,
        model_depth=False,
    )
    vae.train(1, save_best=False)
    vae = PEAKVI(
        data,
        region_factors=False,
    )
    vae.train(1, save_best=False)
    vae = PEAKVI(
        data,
    )
    vae.train(3)
    vae.get_elbo(indices=vae.validation_indices)
    vae.get_accessibility_estimates()
    vae.get_accessibility_estimates(normalize_cells=True)
    vae.get_accessibility_estimates(normalize_regions=True)
    vae.get_library_size_factors()
    vae.get_region_factors()
    vae.get_reconstruction_error(indices=vae.validation_indices)
    vae.get_latent_representation()
    vae.differential_accessibility(groupby="labels", group1="label_1")


def single_pass_for_online_update(model):
    dl = model._make_data_loader(model.adata, indices=range(0, 10))
    for tensors in dl:
        _, _, scvi_loss = model.module(tensors)
    scvi_loss.loss.backward()


def test_peakvi_online_update(save_path):
    n_latent = 5
    adata1 = synthetic_iid()
    PEAKVI.setup_anndata(adata1, batch_key="batch", labels_key="labels")
    model = PEAKVI(adata1, n_latent=n_latent)
    model.train(1, save_best=False)
    dir_path = os.path.join(save_path, "saved_model/")
    model.save(dir_path, overwrite=True)

    # also test subset var option
    adata2 = synthetic_iid()
    adata2.obs["batch"] = adata2.obs.batch.cat.rename_categories(["batch_2", "batch_3"])

    model2 = PEAKVI.load_query_data(adata2, dir_path)
    model2.train(max_epochs=1, weight_decay=0.0, save_best=False)
    model2.get_latent_representation()
    single_pass_for_online_update(model2)

    # encoder linear layer equal for peak features
    one = (
        model.module.z_encoder.encoder.fc_layers[0][0]
        .weight.detach()
        .cpu()
        .numpy()[:, : adata1.shape[1]]
    )
    two = (
        model2.module.z_encoder.encoder.fc_layers[0][0]
        .weight.detach()
        .cpu()
        .numpy()[:, : adata1.shape[1]]
    )
    np.testing.assert_equal(one, two)
    assert (
        np.sum(
            model2.module.z_encoder.encoder.fc_layers[0][0]
            .weight.grad.cpu()
            .numpy()[:, : adata1.shape[1]]
        )
        == 0
    )

    # test options
    adata1 = synthetic_iid()
    PEAKVI.setup_anndata(adata1, batch_key="batch", labels_key="labels")
    model = PEAKVI(
        adata1,
        n_latent=n_latent,
        encode_covariates=True,
    )
    model.train(1, check_val_every_n_epoch=1, save_best=False)
    dir_path = os.path.join(save_path, "saved_model/")
    model.save(dir_path, overwrite=True)

    adata2 = synthetic_iid()
    adata2.obs["batch"] = adata2.obs.batch.cat.rename_categories(["batch_2", "batch_3"])

    model2 = PEAKVI.load_query_data(adata2, dir_path, freeze_expression=True)
    model2.train(max_epochs=1, weight_decay=0.0, save_best=False)
    # deactivate no grad decorator
    model2.get_latent_representation()
    # pytorch lightning zeros the grad, so this will get a grad to inspect
    single_pass_for_online_update(model2)
    grad = model2.module.z_encoder.encoder.fc_layers[0][0].weight.grad.cpu().numpy()
    # expression part has zero grad
    assert np.sum(grad[:, :-4]) == 0
    # categorical part has non-zero grad
    assert np.count_nonzero(grad[:, -4:]) > 0

    # do not freeze expression
    model3 = PEAKVI.load_query_data(
        adata2,
        dir_path,
        freeze_expression=False,
        freeze_decoder_first_layer=False,
    )
    model3.train(max_epochs=1, save_best=False, weight_decay=0.0)
    model3.get_latent_representation()
    single_pass_for_online_update(model3)
    grad = model3.module.z_encoder.encoder.fc_layers[0][0].weight.grad.cpu().numpy()
    # linear layer weight in encoder layer has non-zero grad
    assert np.count_nonzero(grad[:, :-4]) != 0
    grad = model3.module.z_decoder.px_decoder.fc_layers[0][0].weight.grad.cpu().numpy()
    # linear layer weight in decoder layer has non-zero grad
    assert np.count_nonzero(grad[:, :-4]) != 0



import os
import pickle
from typing import Literal

import numpy as np
import torch
from anndata import AnnData, read_h5ad

from scvi.data._download import _download


def _load_legacy_saved_gimvi_files(
    dir_path: str,
    file_name_prefix: str,
    load_seq_adata: bool,
    load_spatial_adata: bool,
) -> tuple[dict, np.ndarray, np.ndarray, dict, AnnData | None, AnnData | None]:
    model_path = os.path.join(dir_path, f"{file_name_prefix}model_params.pt")
    setup_dict_path = os.path.join(dir_path, f"{file_name_prefix}attr.pkl")
    seq_var_names_path = os.path.join(dir_path, f"{file_name_prefix}var_names_seq.csv")
    spatial_var_names_path = os.path.join(dir_path, f"{file_name_prefix}var_names_spatial.csv")

    model_state_dict = torch.load(model_path, map_location="cpu", weights_only=False)

    seq_var_names = np.genfromtxt(seq_var_names_path, delimiter=",", dtype=str)
    spatial_var_names = np.genfromtxt(spatial_var_names_path, delimiter=",", dtype=str)

    with open(setup_dict_path, "rb") as handle:
        attr_dict = pickle.load(handle)

    adata_seq, adata_spatial = None, None
    if load_seq_adata:
        seq_data_path = os.path.join(dir_path, f"{file_name_prefix}adata_seq.h5ad")
        if os.path.exists(seq_data_path):
            adata_seq = read_h5ad(seq_data_path)
        elif not os.path.exists(seq_data_path):
            raise ValueError("Save path contains no saved anndata and no adata was passed.")
    if load_spatial_adata:
        spatial_data_path = os.path.join(dir_path, f"{file_name_prefix}adata_spatial.h5ad")
        if os.path.exists(spatial_data_path):
            adata_spatial = read_h5ad(spatial_data_path)
        elif not os.path.exists(spatial_data_path):
            raise ValueError("Save path contains no saved anndata and no adata was passed.")

    return (
        model_state_dict,
        seq_var_names,
        spatial_var_names,
        attr_dict,
        adata_seq,
        adata_spatial,
    )


def _load_saved_gimvi_files(
    dir_path: str,
    load_seq_adata: bool,
    load_spatial_adata: bool,
    prefix: str | None = None,
    map_location: Literal["cpu", "cuda"] | None = None,
    backup_url: str | None = None,
) -> tuple[dict, dict, np.ndarray, np.ndarray, dict, AnnData | None, AnnData | None]:
    file_name_prefix = prefix or ""

    model_file_name = f"{file_name_prefix}model.pt"
    model_path = os.path.join(dir_path, model_file_name)
    try:
        _download(backup_url, dir_path, model_file_name)
        model = torch.load(model_path, map_location=map_location, weights_only=False)
    except FileNotFoundError as exc:
        raise ValueError(
            f"Failed to load model file at {model_path}. "
            "If attempting to load a saved model from <v0.15.0, please use the util function "
            "`convert_legacy_save` to convert to an updated format."
        ) from exc

    model_state_dict = model["model_state_dict"]
    seq_var_names = model["seq_var_names"]
    spatial_var_names = model["spatial_var_names"]
    attr_dict = model["attr_dict"]

    adata_seq, adata_spatial = None, None
    if load_seq_adata:
        seq_data_path = os.path.join(dir_path, f"{file_name_prefix}adata_seq.h5ad")
        if os.path.exists(seq_data_path):
            adata_seq = read_h5ad(seq_data_path)
        elif not os.path.exists(seq_data_path):
            raise ValueError("Save path contains no saved anndata and no adata was passed.")
    if load_spatial_adata:
        spatial_data_path = os.path.join(dir_path, f"{file_name_prefix}adata_spatial.h5ad")
        if os.path.exists(spatial_data_path):
            adata_spatial = read_h5ad(spatial_data_path)
        elif not os.path.exists(spatial_data_path):
            raise ValueError("Save path contains no saved anndata and no adata was passed.")

    return (
        attr_dict,
        seq_var_names,
        spatial_var_names,
        model_state_dict,
        adata_seq,
        adata_spatial,
    )



import logging
import warnings
from collections.abc import Iterable as IterableClass

import anndata
import numpy as np
import pandas as pd

from scvi import settings
from scvi.data._constants import _ADATA_MINIFY_TYPE_UNS_KEY, ADATA_MINIFY_TYPE
from scvi.utils import track

from ._differential import DifferentialComputation

logger = logging.getLogger(__name__)


def _prepare_obs(
    idx1: list[bool] | np.ndarray | str,
    idx2: list[bool] | np.ndarray | str,
    adata: anndata.AnnData,
):
    """Construct an array used for masking.

    Given population identifiers `idx1` and potentially `idx2`,
    this function creates an array `obs_col` that identifies both populations
    for observations contained in `adata`.
    In particular, `obs_col` will take values `group1` (resp. `group2`)
    for `idx1` (resp `idx2`).

    Parameters
    ----------
    idx1
        Can be of three types. First, it can corresponds to a boolean mask that
        has the same shape as adata. It can also corresponds to a list of indices.
        Last, it can correspond to string query of adata.obs columns.
    idx2
        Same as above
    adata
        Anndata
    """

    def ravel_idx(my_idx, obs_df):
        return (
            obs_df.index.isin(obs_df.query(my_idx).index)
            if isinstance(my_idx, str)
            else np.asarray(my_idx).ravel()
        )

    obs_df = adata.obs
    idx1 = ravel_idx(idx1, obs_df)
    g1_key = "one"
    obs_col = np.array(["None"] * adata.shape[0], dtype=str)
    obs_col[idx1] = g1_key
    group1 = [g1_key]
    group2 = None if idx2 is None else "two"
    if idx2 is not None:
        idx2 = ravel_idx(idx2, obs_df)
        obs_col[idx2] = group2
    if (obs_col[idx1].shape[0] == 0) or (obs_col[idx2].shape[0] == 0):
        raise ValueError("One of idx1 or idx2 has size zero.")
    return obs_col, group1, group2


def _de_core(
    adata_manager,
    model_fn,
    representation_fn,
    groupby,
    group1,
    group2,
    idx1,
    idx2,
    all_stats,
    all_stats_fn,
    col_names,
    mode,
    batchid1,
    batchid2,
    delta,
    batch_correction,
    fdr,
    silent,
    **kwargs,
):
    """Internal function for DE interface."""
    if (
        adata_manager.adata.uns.get(_ADATA_MINIFY_TYPE_UNS_KEY, None)
        == ADATA_MINIFY_TYPE.LATENT_POSTERIOR
    ):
        warnings.warn(
            "Count statistics make no sense for minified model. Consider disabling all_stats.",
            UserWarning,
            stacklevel=settings.warnings_stacklevel,
        )
    adata = adata_manager.adata
    if group1 is None and idx1 is None:
        group1 = adata.obs[groupby].astype("category").cat.categories.tolist()
        if len(group1) == 1:
            raise ValueError("Only a single group in the data. Can't run DE on a single group.")

    if not isinstance(group1, IterableClass) or isinstance(group1, str):
        group1 = [group1]

    # make a temp obs key using indices
    temp_key = None
    if idx1 is not None:
        obs_col, group1, group2 = _prepare_obs(idx1, idx2, adata)
        temp_key = "_scvi_temp_de"
        adata.obs[temp_key] = obs_col
        groupby = temp_key

    df_results = []
    dc = DifferentialComputation(model_fn, representation_fn, adata_manager)
    for g1 in track(
        group1,
        description="DE...",
        disable=silent,
    ):
        cell_idx1 = (adata.obs[groupby] == g1).to_numpy().ravel()
        if group2 is None:
            cell_idx2 = ~cell_idx1
        else:
            cell_idx2 = (adata.obs[groupby] == group2).to_numpy().ravel()

        all_info = dc.get_bayes_factors(
            cell_idx1,
            cell_idx2,
            mode=mode,
            delta=delta,
            batchid1=batchid1,
            batchid2=batchid2,
            use_observed_batches=not batch_correction,
            **kwargs,
        )

        if all_stats is True:
            genes_properties_dict = all_stats_fn(adata_manager, cell_idx1, cell_idx2)
            all_info = {**all_info, **genes_properties_dict}

        res = pd.DataFrame(all_info, index=col_names)
        sort_key = "proba_de" if mode == "change" else "bayes_factor"
        res = res.sort_values(by=sort_key, ascending=False)
        if mode == "change":
            res[f"is_de_fdr_{fdr}"] = _fdr_de_prediction(res["proba_de"], fdr=fdr)
        if idx1 is None:
            g2 = "Rest" if group2 is None else group2
            res["comparison"] = f"{g1} vs {g2}"
            res["group1"] = g1
            res["group2"] = g2
        df_results.append(res)

    if temp_key is not None:
        del adata.obs[temp_key]

    result = pd.concat(df_results, axis=0)

    return result


def _fdr_de_prediction(posterior_probas: pd.Series, fdr: float = 0.05) -> pd.Series:
    """Compute posterior expected FDR and tag features as DE."""
    if not posterior_probas.ndim == 1:
        raise ValueError("posterior_probas should be 1-dimensional")
    original_index = posterior_probas.index
    sorted_pgs = posterior_probas.sort_values(ascending=False)
    cumulative_fdr = (1.0 - sorted_pgs).cumsum() / (1.0 + np.arange(len(sorted_pgs)))
    d = (cumulative_fdr <= fdr).sum()
    is_pred_de = pd.Series(np.zeros_like(cumulative_fdr).astype(bool), index=sorted_pgs.index)
    is_pred_de.iloc[:d] = True
    is_pred_de = is_pred_de.loc[original_index]
    return is_pred_de



from __future__ import annotations

import logging
from typing import TYPE_CHECKING

import numpy as np
import pandas as pd
import torch

from scvi import REGISTRY_KEYS
from scvi.data import AnnDataManager
from scvi.data.fields import LayerField, NumericalObsField
from scvi.model.base import BaseModelClass, UnsupervisedTrainingMixin
from scvi.module import MRDeconv
from scvi.utils import setup_anndata_dsp
from scvi.utils._docstrings import devices_dsp

if TYPE_CHECKING:
    from collections import OrderedDict
    from collections.abc import Sequence

    from anndata import AnnData

    from scvi.model import CondSCVI

logger = logging.getLogger(__name__)


class DestVI(UnsupervisedTrainingMixin, BaseModelClass):
    """Multi-resolution deconvolution of Spatial Transcriptomics data (DestVI) :cite:p:`Lopez22`.

    Most users will use the alternate constructor (see example).

    Parameters
    ----------
    st_adata
        spatial transcriptomics AnnData object that has been registered via
        :meth:`~scvi.model.DestVI.setup_anndata`.
    cell_type_mapping
        mapping between numerals and cell type labels
    decoder_state_dict
        state_dict from the decoder of the CondSCVI model
    px_decoder_state_dict
        state_dict from the px_decoder of the CondSCVI model
    px_r
        parameters for the px_r tensor in the CondSCVI model
    n_hidden
        Number of nodes per hidden layer.
    n_latent
        Dimensionality of the latent space.
    n_layers
        Number of hidden layers used for encoder and decoder NNs.
    **module_kwargs
        Keyword args for :class:`~scvi.modules.MRDeconv`

    Examples
    --------
    >>> sc_adata = anndata.read_h5ad(path_to_scRNA_anndata)
    >>> scvi.model.CondSCVI.setup_anndata(sc_adata)
    >>> sc_model = scvi.model.CondSCVI(sc_adata)
    >>> st_adata = anndata.read_h5ad(path_to_ST_anndata)
    >>> DestVI.setup_anndata(st_adata)
    >>> spatial_model = DestVI.from_rna_model(st_adata, sc_model)
    >>> spatial_model.train(max_epochs=2000)
    >>> st_adata.obsm["proportions"] = spatial_model.get_proportions(st_adata)
    >>> gamma = spatial_model.get_gamma(st_adata)

    Notes
    -----
    See further usage examples in the following tutorials:

    1. :doc:`/tutorials/notebooks/spatial/DestVI_tutorial`
    2. :doc:`/tutorials/notebooks/spatial/DestVI_in_R`
    """

    _module_cls = MRDeconv

    def __init__(
        self,
        st_adata: AnnData,
        cell_type_mapping: np.ndarray,
        decoder_state_dict: OrderedDict,
        px_decoder_state_dict: OrderedDict,
        px_r: np.ndarray,
        n_hidden: int,
        n_latent: int,
        n_layers: int,
        dropout_decoder: float,
        l1_reg: float,
        **module_kwargs,
    ):
        super().__init__(st_adata)
        self.module = self._module_cls(
            n_spots=st_adata.n_obs,
            n_labels=cell_type_mapping.shape[0],
            decoder_state_dict=decoder_state_dict,
            px_decoder_state_dict=px_decoder_state_dict,
            px_r=px_r,
            n_genes=st_adata.n_vars,
            n_latent=n_latent,
            n_layers=n_layers,
            n_hidden=n_hidden,
            dropout_decoder=dropout_decoder,
            l1_reg=l1_reg,
            **module_kwargs,
        )
        self.cell_type_mapping = cell_type_mapping
        self._model_summary_string = "DestVI Model"
        self.init_params_ = self._get_init_params(locals())

    @classmethod
    def from_rna_model(
        cls,
        st_adata: AnnData,
        sc_model: CondSCVI,
        vamp_prior_p: int = 15,
        l1_reg: float = 0.0,
        **module_kwargs,
    ):
        """Alternate constructor for exploiting a pre-trained model on a RNA-seq dataset.

        Parameters
        ----------
        st_adata
            registered anndata object
        sc_model
            trained CondSCVI model
        vamp_prior_p
            number of mixture parameter for VampPrior calculations
        l1_reg
            Scalar parameter indicating the strength of L1 regularization on cell type proportions.
            A value of 50 leads to sparser results.
        **model_kwargs
            Keyword args for :class:`~scvi.model.DestVI`
        """
        decoder_state_dict = sc_model.module.decoder.state_dict()
        px_decoder_state_dict = sc_model.module.px_decoder.state_dict()
        px_r = sc_model.module.px_r.detach().cpu().numpy()
        mapping = sc_model.adata_manager.get_state_registry(
            REGISTRY_KEYS.LABELS_KEY
        ).categorical_mapping
        dropout_decoder = sc_model.module.dropout_rate
        if vamp_prior_p is None:
            mean_vprior = None
            var_vprior = None
        else:
            mean_vprior, var_vprior, mp_vprior = sc_model.get_vamp_prior(
                sc_model.adata, p=vamp_prior_p
            )

        return cls(
            st_adata,
            mapping,
            decoder_state_dict,
            px_decoder_state_dict,
            px_r,
            sc_model.module.n_hidden,
            sc_model.module.n_latent,
            sc_model.module.n_layers,
            mean_vprior=mean_vprior,
            var_vprior=var_vprior,
            mp_vprior=mp_vprior,
            dropout_decoder=dropout_decoder,
            l1_reg=l1_reg,
            **module_kwargs,
        )

    def get_proportions(
        self,
        keep_noise: bool = False,
        indices: Sequence[int] | None = None,
        batch_size: int | None = None,
    ) -> pd.DataFrame:
        """Returns the estimated cell type proportion for the spatial data.

        Shape is n_cells x n_labels OR n_cells x (n_labels + 1) if keep_noise.

        Parameters
        ----------
        keep_noise
            whether to account for the noise term as a standalone cell type in the proportion
            estimate.
        indices
            Indices of cells in adata to use. Only used if amortization. If `None`, all cells are
            used.
        batch_size
            Minibatch size for data loading into model. Only used if amortization. Defaults to
            `scvi.settings.batch_size`.
        """
        self._check_if_trained()

        column_names = self.cell_type_mapping
        index_names = self.adata.obs.index
        if keep_noise:
            column_names = np.append(column_names, "noise_term")

        if self.module.amortization in ["both", "proportion"]:
            stdl = self._make_data_loader(adata=self.adata, indices=indices, batch_size=batch_size)
            prop_ = []
            for tensors in stdl:
                generative_inputs = self.module._get_generative_input(tensors, None)
                prop_local = self.module.get_proportions(
                    x=generative_inputs["x"], keep_noise=keep_noise
                )
                prop_ += [prop_local.cpu()]
            data = torch.cat(prop_).numpy()
            if indices:
                index_names = index_names[indices]
        else:
            if indices is not None:
                logger.info(
                    "No amortization for proportions, ignoring indices and returning results for "
                    "the full data"
                )
            data = self.module.get_proportions(keep_noise=keep_noise)

        return pd.DataFrame(
            data=data,
            columns=column_names,
            index=index_names,
        )

    def get_gamma(
        self,
        indices: Sequence[int] | None = None,
        batch_size: int | None = None,
        return_numpy: bool = False,
    ) -> np.ndarray | dict[str, pd.DataFrame]:
        """Returns the estimated cell-type specific latent space for the spatial data.

        Parameters
        ----------
        indices
            Indices of cells in adata to use. Only used if amortization. If `None`, all cells are
            used.
        batch_size
            Minibatch size for data loading into model. Only used if amortization. Defaults to
            `scvi.settings.batch_size`.
        return_numpy
            if activated, will return a numpy array of shape is n_spots x n_latent x n_labels.
        """
        self._check_if_trained()

        column_names = np.arange(self.module.n_latent)
        index_names = self.adata.obs.index

        if self.module.amortization in ["both", "latent"]:
            stdl = self._make_data_loader(adata=self.adata, indices=indices, batch_size=batch_size)
            gamma_ = []
            for tensors in stdl:
                generative_inputs = self.module._get_generative_input(tensors, None)
                gamma_local = self.module.get_gamma(x=generative_inputs["x"])
                gamma_ += [gamma_local.cpu()]
            data = torch.cat(gamma_, dim=-1).numpy()
            if indices is not None:
                index_names = index_names[indices]
        else:
            if indices is not None:
                logger.info(
                    "No amortization for latent values, ignoring adata and returning results for "
                    "the full data"
                )
            data = self.module.get_gamma()

        data = np.transpose(data, (2, 0, 1))
        if return_numpy:
            return data
        else:
            res = {}
            for i, ct in enumerate(self.cell_type_mapping):
                res[ct] = pd.DataFrame(data=data[:, :, i], columns=column_names, index=index_names)
            return res

    def get_scale_for_ct(
        self,
        label: str,
        indices: Sequence[int] | None = None,
        batch_size: int | None = None,
    ) -> pd.DataFrame:
        r"""Return the scaled parameter of the NB for every spot in queried cell types.

        Parameters
        ----------
        label
            cell type of interest
        indices
            Indices of cells in self.adata to use. If `None`, all cells are used.
        batch_size
            Minibatch size for data loading into model. Defaults to `scvi.settings.batch_size`.

        Returns
        -------
        Pandas dataframe of gene_expression
        """
        self._check_if_trained()

        if label not in self.cell_type_mapping:
            raise ValueError("Unknown cell type")
        y = np.where(label == self.cell_type_mapping)[0][0]

        stdl = self._make_data_loader(self.adata, indices=indices, batch_size=batch_size)
        scale = []
        for tensors in stdl:
            generative_inputs = self.module._get_generative_input(tensors, None)
            x, ind_x = (
                generative_inputs["x"],
                generative_inputs["ind_x"],
            )
            px_scale = self.module.get_ct_specific_expression(x, ind_x, y)
            scale += [px_scale.cpu()]

        data = torch.cat(scale).numpy()
        column_names = self.adata.var.index
        index_names = self.adata.obs.index
        if indices is not None:
            index_names = index_names[indices]
        return pd.DataFrame(data=data, columns=column_names, index=index_names)

    @devices_dsp.dedent
    def train(
        self,
        max_epochs: int = 2000,
        lr: float = 0.003,
        accelerator: str = "auto",
        devices: int | list[int] | str = "auto",
        train_size: float = 1.0,
        validation_size: float | None = None,
        shuffle_set_split: bool = True,
        batch_size: int = 128,
        n_epochs_kl_warmup: int = 200,
        datasplitter_kwargs: dict | None = None,
        plan_kwargs: dict | None = None,
        **kwargs,
    ):
        """Trains the model using MAP inference.

        Parameters
        ----------
        max_epochs
            Number of epochs to train for
        lr
            Learning rate for optimization.
        %(param_accelerator)s
        %(param_devices)s
        train_size
            Size of training set in the range [0.0, 1.0].
        validation_size
            Size of the test set. If `None`, defaults to 1 - `train_size`. If
            `train_size + validation_size < 1`, the remaining cells belong to a test set.
        shuffle_set_split
            Whether to shuffle indices before splitting. If `False`, the val, train, and test set
            are split in the sequential order of the data according to `validation_size` and
            `train_size` percentages.
        batch_size
            Minibatch size to use during training.
        n_epochs_kl_warmup
            number of epochs needed to reach unit kl weight in the elbo
        datasplitter_kwargs
            Additional keyword arguments passed into :class:`~scvi.dataloaders.DataSplitter`.
        plan_kwargs
            Keyword args for :class:`~scvi.train.TrainingPlan`. Keyword arguments passed to
            `train()` will overwrite values present in `plan_kwargs`, when appropriate.
        **kwargs
            Other keyword args for :class:`~scvi.train.Trainer`.
        """
        update_dict = {
            "lr": lr,
            "n_epochs_kl_warmup": n_epochs_kl_warmup,
        }
        if plan_kwargs is not None:
            plan_kwargs.update(update_dict)
        else:
            plan_kwargs = update_dict
        super().train(
            max_epochs=max_epochs,
            accelerator=accelerator,
            devices=devices,
            train_size=train_size,
            validation_size=validation_size,
            shuffle_set_split=shuffle_set_split,
            batch_size=batch_size,
            datasplitter_kwargs=datasplitter_kwargs,
            plan_kwargs=plan_kwargs,
            **kwargs,
        )

    @classmethod
    @setup_anndata_dsp.dedent
    def setup_anndata(
        cls,
        adata: AnnData,
        layer: str | None = None,
        **kwargs,
    ):
        """%(summary)s.

        Parameters
        ----------
        %(param_adata)s
        %(param_layer)s
        """
        setup_method_args = cls._get_setup_method_args(**locals())
        # add index for each cell (provided to pyro plate for correct minibatching)
        adata.obs["_indices"] = np.arange(adata.n_obs)
        anndata_fields = [
            LayerField(REGISTRY_KEYS.X_KEY, layer, is_count_data=True),
            NumericalObsField(REGISTRY_KEYS.INDICES_KEY, "_indices"),
        ]
        adata_manager = AnnDataManager(fields=anndata_fields, setup_method_args=setup_method_args)
        adata_manager.register_fields(adata, **kwargs)
        cls.register_manager(adata_manager)



import scipy

from scvi.data import synthetic_iid
from scvi.external import SCAR


def test_scar():
    n_latent = 5
    adata = synthetic_iid()

    adata.obs["batch"] = adata.obs["batch"].cat.rename_categories(["batch_2", "batch_3"])

    adata.X = scipy.sparse.csr_matrix(adata.X)
    SCAR.setup_anndata(adata, batch_key="batch", labels_key="labels")

    _ = SCAR.get_ambient_profile(adata, adata, prob=0.0, iterations=1, sample=100)
    model = SCAR(adata, ambient_profile=None, n_latent=n_latent)
    model.train(1, check_val_every_n_epoch=1, train_size=0.5)
    model.get_elbo()
    model.get_latent_representation()
    model.get_marginal_ll(n_mc_samples=5)
    model.get_reconstruction_error()
    model.get_denoised_counts(adata, n_samples=1)
    _ = model.history

    # tests __repr__
    print(model)



from __future__ import annotations

import warnings
from typing import TYPE_CHECKING

import flax.linen as nn
import jax
import jax.numpy as jnp
import numpyro.distributions as dist

from scvi import REGISTRY_KEYS, settings
from scvi.distributions import JaxNegativeBinomialMeanDisp as NegativeBinomial
from scvi.external.mrvi._components import AttentionBlock, Dense
from scvi.module.base import JaxBaseModuleClass, LossOutput, flax_configure

if TYPE_CHECKING:
    from collections.abc import Callable
    from typing import Any

DEFAULT_PX_KWARGS = {
    "n_hidden": 32,
    "stop_gradients": False,
    "stop_gradients_mlp": True,
    "dropout_rate": 0.03,
}
DEFAULT_QZ_ATTENTION_KWARGS = {
    "use_map": True,
    "stop_gradients": False,
    "stop_gradients_mlp": True,
    "dropout_rate": 0.03,
}
DEFAULT_QU_KWARGS = {}

# Lower stddev leads to better initial loss values
_normal_initializer = jax.nn.initializers.normal(stddev=0.1)


class DecoderZXAttention(nn.Module):
    """Attention-based decoder.

    Parameters
    ----------
    n_in
        Number of input features.
    n_out
        Number of output features.
    n_batch
        Number of batches.
    n_latent_sample
        Number of latent samples.
    h_activation
        Activation function for the output layer.
    n_channels
        Number of channels in the attention block.
    n_heads
        Number of heads in the attention block.
    dropout_rate
        Dropout rate.
    stop_gradients
        Whether to stop gradients to ``z``.
    stop_gradients_mlp
        Whether to stop gradients to the MLP in the attention block.
    training
        Whether the model is in training mode.
    n_hidden
        Number of hidden units in the MLP.
    n_layers
        Number of layers in the MLP.
    low_dim_batch
        Whether to use low-dimensional batch embeddings.
    activation
        Activation function for the MLP.
    """

    n_in: int
    n_out: int
    n_batch: int
    n_latent_sample: int = 16
    h_activation: Callable[[jax.typing.ArrayLike], jax.Array] = nn.softmax
    n_channels: int = 4
    n_heads: int = 2
    dropout_rate: float = 0.1
    stop_gradients: bool = False
    stop_gradients_mlp: bool = False
    training: bool | None = None
    n_hidden: int = 32
    n_layers: int = 1
    training: bool | None = None
    low_dim_batch: bool = True
    activation: Callable[[jax.typing.ArrayLike], jax.Array] = nn.gelu

    @nn.compact
    def __call__(
        self,
        z: jax.typing.ArrayLike,
        batch_covariate: jax.typing.ArrayLike,
        size_factor: jax.typing.ArrayLike,
        training: bool | None = None,
    ) -> NegativeBinomial:
        has_mc_samples = z.ndim == 3
        z_stop = z if not self.stop_gradients else jax.lax.stop_gradient(z)
        z_ = nn.LayerNorm(name="u_ln")(z_stop)

        batch_covariate = batch_covariate.astype(int).flatten()

        if self.n_batch >= 2:
            batch_embed = nn.Embed(
                self.n_batch, self.n_latent_sample, embedding_init=_normal_initializer
            )(batch_covariate)  # (batch, n_latent_sample)
            batch_embed = nn.LayerNorm(name="batch_embed_ln")(batch_embed)
            if has_mc_samples:
                batch_embed = jnp.tile(batch_embed, (z_.shape[0], 1, 1))

            res_dim = self.n_in if self.low_dim_batch else self.n_out

            query_embed = z_
            kv_embed = batch_embed
            residual = AttentionBlock(
                query_dim=self.n_in,
                out_dim=res_dim,
                outerprod_dim=self.n_latent_sample,
                n_channels=self.n_channels,
                n_heads=self.n_heads,
                dropout_rate=self.dropout_rate,
                n_hidden_mlp=self.n_hidden,
                n_layers_mlp=self.n_layers,
                stop_gradients_mlp=self.stop_gradients_mlp,
                training=training,
                activation=self.activation,
            )(query_embed=query_embed, kv_embed=kv_embed)

            if self.low_dim_batch:
                mu = nn.Dense(self.n_out)(z + residual)
            else:
                mu = nn.Dense(self.n_out)(z) + residual
        else:
            mu = nn.Dense(self.n_out)(z_)
        mu = self.h_activation(mu)
        return NegativeBinomial(
            mean=mu * size_factor,
            inverse_dispersion=jnp.exp(self.param("px_r", jax.random.normal, (self.n_out,))),
        )


class EncoderUZ(nn.Module):
    """Attention-based encoder from ``u`` to ``z``.

    Parameters
    ----------
    n_latent
        Number of latent variables.
    n_sample
        Number of samples.
    n_latent_u
        Number of latent variables for ``u``.
    n_latent_sample
        Number of latent samples.
    n_channels
        Number of channels in the attention block.
    n_heads
        Number of heads in the attention block.
    dropout_rate
        Dropout rate.
    stop_gradients
        Whether to stop gradients to ``u``.
    stop_gradients_mlp
        Whether to stop gradients to the MLP in the attention block.
    use_map
        Whether to use the MAP estimate to approximate the posterior of ``z`` given ``u``
    n_hidden
        Number of hidden units in the MLP.
    n_layers
        Number of layers in the MLP.
    training
        Whether the model is in training mode.
    activation
        Activation function for the MLP.
    """

    n_latent: int
    n_sample: int
    n_latent_u: int | None = None
    n_latent_sample: int = 16
    n_channels: int = 4
    n_heads: int = 2
    dropout_rate: float = 0.0
    stop_gradients: bool = False
    stop_gradients_mlp: bool = False
    use_map: bool = True
    n_hidden: int = 32
    n_layers: int = 1
    training: bool | None = None
    activation: Callable[[jax.typing.ArrayLike], jax.Array] = nn.gelu

    @nn.compact
    def __call__(
        self,
        u: jax.typing.ArrayLike,
        sample_covariate: jax.typing.ArrayLike,
        training: bool | None = None,
    ) -> tuple[jax.Array, jax.Array]:
        training = nn.merge_param("training", self.training, training)
        sample_covariate = sample_covariate.astype(int).flatten()
        self.n_latent_u if self.n_latent_u is not None else self.n_latent  # noqa: B018
        has_mc_samples = u.ndim == 3
        u_stop = u if not self.stop_gradients else jax.lax.stop_gradient(u)
        u_ = nn.LayerNorm(name="u_ln")(u_stop)

        sample_embed = nn.Embed(
            self.n_sample, self.n_latent_sample, embedding_init=_normal_initializer
        )(sample_covariate)  # (batch, n_latent_sample)
        sample_embed = nn.LayerNorm(name="sample_embed_ln")(sample_embed)
        if has_mc_samples:
            sample_embed = jnp.tile(sample_embed, (u_.shape[0], 1, 1))

        n_outs = 1 if self.use_map else 2
        residual = AttentionBlock(
            query_dim=self.n_latent,
            out_dim=n_outs * self.n_latent,
            outerprod_dim=self.n_latent_sample,
            n_channels=self.n_channels,
            n_heads=self.n_heads,
            dropout_rate=self.dropout_rate,
            stop_gradients_mlp=self.stop_gradients_mlp,
            n_hidden_mlp=self.n_hidden,
            n_layers_mlp=self.n_layers,
            training=training,
            activation=self.activation,
        )(query_embed=u_, kv_embed=sample_embed)

        if self.n_latent_u is not None:
            z_base = nn.Dense(self.n_latent)(u_stop)
            return z_base, residual
        else:
            return u, residual


class EncoderXU(nn.Module):
    """Encoder from ``x`` to ``u``.

    Parameters
    ----------
    n_latent
        Number of latent variables.
    n_sample
        Number of samples.
    n_hidden
        Number of hidden units in the MLP.
    n_layers
        Number of layers in the MLP.
    activation
        Activation function for the MLP.
    training
        Whether the model is in training mode.
    """

    n_latent: int
    n_sample: int
    n_hidden: int
    n_layers: int = 1
    activation: Callable[[jax.typing.ArrayLike], jax.Array] = nn.gelu
    training: bool | None = None

    @nn.compact
    def __call__(
        self,
        x: jax.typing.ArrayLike,
        sample_covariate: jax.typing.ArrayLike,
        training: bool | None = None,
    ) -> dist.Normal:
        from scvi.external.mrvi._components import ConditionalNormalization, NormalDistOutputNN

        training = nn.merge_param("training", self.training, training)
        x_feat = jnp.log1p(x)
        for _ in range(2):
            x_feat = Dense(self.n_hidden)(x_feat)
            x_feat = ConditionalNormalization(self.n_hidden, self.n_sample)(
                x_feat, sample_covariate, training=training
            )
            x_feat = self.activation(x_feat)
        sample_effect = nn.Embed(self.n_sample, self.n_hidden, embedding_init=_normal_initializer)(
            sample_covariate.squeeze(-1).astype(int)
        )
        inputs = x_feat + sample_effect
        return NormalDistOutputNN(self.n_latent, self.n_hidden, self.n_layers)(
            inputs, training=training
        )


@flax_configure
class MRVAE(JaxBaseModuleClass):
    """Multi-resolution Variational Inference (MrVI) module.

    Parameters
    ----------
    n_input
        Number of input features.
    n_sample
        Number of samples.
    n_batch
        Number of batches.
    n_labels
        Number of labels.
    n_latent
        Number of latent variables for ``z``.
    n_latent_u
        Number of latent variables for ``u``.
    encoder_n_hidden
        Number of hidden units in the encoder.
    encoder_n_layers
        Number of layers in the encoder.
    z_u_prior
        Whether to place a Gaussian prior on ``z`` given ``u``.
    z_u_prior_scale
        Natural log of the scale parameter of the Gaussian prior placed on ``z`` given ``u``. Only
        applies of ``learn_z_u_prior_scale`` is ``False``.
    u_prior_scale
        Natural log of the scale parameter of the Gaussian prior placed on ``u``. If
        ``u_prior_mixture`` is ``True``, this scale applies to each mixture component distribution.
    u_prior_mixture
        Whether to use a mixture of Gaussians prior for ``u``.
    u_prior_mixture_k
        Number of mixture components to use for the mixture of Gaussians prior on ``u``.
    learn_z_u_prior_scale
        Whether to learn the scale parameter of the prior distribution of ``z`` given ``u``.
    scale_observations
        Whether to scale the loss associated with each observation by the total number of
        observations linked to the associated sample.
    px_kwargs
        Keyword arguments for the generative model.
    qz_kwargs
        Keyword arguments for the inference model from ``u`` to ``z``.
    qu_kwargs
        Keyword arguments for the inference model from ``x`` to ``u``.
    training
        Whether the model is in training mode.
    n_obs_per_sample
        Number of observations per sample.
    """

    n_input: int
    n_sample: int
    n_batch: int
    n_labels: int
    n_latent: int = 30
    n_latent_u: int = 10
    encoder_n_hidden: int = 128
    encoder_n_layers: int = 2
    z_u_prior: bool = True
    z_u_prior_scale: float = 0.0
    u_prior_scale: float = 0.0
    u_prior_mixture: bool = True
    u_prior_mixture_k: int = 20
    learn_z_u_prior_scale: bool = False
    scale_observations: bool = False
    px_kwargs: dict | None = None
    qz_kwargs: dict | None = None
    qu_kwargs: dict | None = None
    training: bool = True
    n_obs_per_sample: jax.typing.ArrayLike | None = None

    def setup(self):
        px_kwargs = DEFAULT_PX_KWARGS.copy()
        if self.px_kwargs is not None:
            px_kwargs.update(self.px_kwargs)

        qz_kwargs = DEFAULT_QZ_ATTENTION_KWARGS.copy()
        if self.qz_kwargs is not None:
            qz_kwargs.update(self.qz_kwargs)

        qu_kwargs = DEFAULT_QU_KWARGS.copy()
        if self.qu_kwargs is not None:
            qu_kwargs.update(self.qu_kwargs)

        is_isomorphic_uz = self.n_latent == self.n_latent_u
        n_latent_u = None if is_isomorphic_uz else self.n_latent_u

        if self.n_latent < self.n_latent_u:
            warnings.warn(
                "The number of latent variables for `z` is set to less than the number of latent "
                "variables for `u`.",
                UserWarning,
                stacklevel=settings.warnings_stacklevel,
            )

        # Generative model
        px_cls = DecoderZXAttention
        self.px = px_cls(
            self.n_latent,
            self.n_input,
            self.n_batch,
            **px_kwargs,
        )

        qz_cls = EncoderUZ
        self.qz = qz_cls(
            self.n_latent,
            self.n_sample,
            n_latent_u=n_latent_u,
            **qz_kwargs,
        )

        # Inference model
        self.qu = EncoderXU(
            n_latent=self.n_latent if is_isomorphic_uz else n_latent_u,
            n_sample=self.n_sample,
            n_hidden=self.encoder_n_hidden,
            n_layers=self.encoder_n_layers,
            **qu_kwargs,
        )

        if self.learn_z_u_prior_scale:
            self.pz_scale = self.param("pz_scale", nn.initializers.zeros, (self.n_latent,))
        else:
            self.pz_scale = self.z_u_prior_scale

        if self.u_prior_mixture:
            if self.n_labels > 1:
                u_prior_mixture_k = self.n_labels
            else:
                u_prior_mixture_k = self.u_prior_mixture_k
            u_dim = self.n_latent_u if self.n_latent_u is not None else self.n_latent
            self.u_prior_logits = self.param(
                "u_prior_logits", nn.initializers.zeros, (u_prior_mixture_k,)
            )
            self.u_prior_means = self.param(
                "u_prior_means", jax.random.normal, (u_dim, u_prior_mixture_k)
            )
            self.u_prior_scales = self.param(
                "u_prior_scales", nn.initializers.zeros, (u_dim, u_prior_mixture_k)
            )

    @property
    def required_rngs(self):
        return ("params", "u", "dropout", "eps")

    def _get_inference_input(self, tensors: dict[str, jax.typing.ArrayLike]) -> dict[str, Any]:
        x = tensors[REGISTRY_KEYS.X_KEY]
        sample_index = tensors[REGISTRY_KEYS.SAMPLE_KEY]
        return {"x": x, "sample_index": sample_index}

    def inference(
        self,
        x: jax.typing.ArrayLike,
        sample_index: jax.typing.ArrayLike,
        mc_samples: int | None = None,
        cf_sample: jax.typing.ArrayLike | None = None,
        use_mean: bool = False,
    ) -> dict[str, jax.Array | dist.Distribution]:
        """Latent variable inference."""
        qu = self.qu(x, sample_index, training=self.training)
        if use_mean:
            u = qu.mean
        else:
            u_rng = self.make_rng("u")
            sample_shape = (mc_samples,) if mc_samples is not None else ()
            u = qu.rsample(u_rng, sample_shape=sample_shape)

        sample_index_cf = sample_index if cf_sample is None else cf_sample

        z_base, eps = self.qz(u, sample_index_cf, training=self.training)
        qeps_ = eps

        qeps = None
        if qeps_.shape[-1] == 2 * self.n_latent:
            loc_, scale_ = qeps_[..., : self.n_latent], qeps_[..., self.n_latent :]
            qeps = dist.Normal(loc_, nn.softplus(scale_) + 1e-3)
            eps = qeps.mean if use_mean else qeps.rsample(self.make_rng("eps"))
        z = z_base + eps
        library = jnp.log(x.sum(1, keepdims=True))

        return {
            "qu": qu,
            "qeps": qeps,
            "eps": eps,
            "u": u,
            "z": z,
            "z_base": z_base,
            "library": library,
        }

    def _get_generative_input(
        self,
        tensors: dict[str, jax.typing.ArrayLike],
        inference_outputs: dict[str, jax.Array | dist.Distribution],
    ) -> dict[str, jax.Array]:
        z = inference_outputs["z"]
        library = inference_outputs["library"]
        batch_index = tensors[REGISTRY_KEYS.BATCH_KEY]
        label_index = tensors[REGISTRY_KEYS.LABELS_KEY]
        return {
            "z": z,
            "library": library,
            "batch_index": batch_index,
            "label_index": label_index,
        }

    def generative(
        self,
        z: jax.typing.ArrayLike,
        library: jax.typing.ArrayLike,
        batch_index: jax.typing.ArrayLike,
        label_index: jax.typing.ArrayLike,
    ) -> dict[str, jax.Array | dist.Distribution]:
        """Generative model."""
        library_exp = jnp.exp(library)
        px = self.px(
            z,
            batch_index,
            size_factor=library_exp,
            training=self.training,
        )
        h = px.mean / library_exp

        if self.u_prior_mixture:
            offset = (
                10.0 * jax.nn.one_hot(label_index, self.n_labels) if self.n_labels >= 2 else 0.0
            )
            cats = dist.Categorical(logits=self.u_prior_logits + offset)
            normal_dists = dist.Normal(self.u_prior_means, jnp.exp(self.u_prior_scales))
            pu = dist.MixtureSameFamily(cats, normal_dists)
        else:
            pu = dist.Normal(0, jnp.exp(self.u_prior_scale))
        return {"px": px, "pu": pu, "h": h}

    def loss(
        self,
        tensors: dict[str, jax.typing.ArrayLike],
        inference_outputs: dict[str, jax.Array | dist.Distribution],
        generative_outputs: dict[str, jax.Array | dist.Distribution],
        kl_weight: float = 1.0,
    ) -> LossOutput:
        """Compute the loss function value."""
        reconstruction_loss = (
            -generative_outputs["px"].log_prob(tensors[REGISTRY_KEYS.X_KEY]).sum(-1)
        )

        if self.u_prior_mixture:
            kl_u = inference_outputs["qu"].log_prob(inference_outputs["u"]) - generative_outputs[
                "pu"
            ].log_prob(inference_outputs["u"])
            kl_u = kl_u.sum(-1)
        else:
            kl_u = dist.kl_divergence(inference_outputs["qu"], generative_outputs["pu"]).sum(-1)
        inference_outputs["qeps"]

        kl_z = 0.0
        eps = inference_outputs["z"] - inference_outputs["z_base"]
        if self.z_u_prior:
            peps = dist.Normal(0, jnp.exp(self.pz_scale))
            kl_z = -peps.log_prob(eps).sum(-1)

        weighted_kl_local = kl_weight * (kl_u + kl_z)
        loss = reconstruction_loss + weighted_kl_local

        if self.scale_observations:
            sample_index = tensors[REGISTRY_KEYS.SAMPLE_KEY].flatten().astype(int)
            prefactors = self.n_obs_per_sample[sample_index]
            loss = loss / prefactors

        loss = jnp.mean(loss)

        return LossOutput(
            loss=loss,
            reconstruction_loss=reconstruction_loss,
            kl_local=(kl_u + kl_z),
        )

    def compute_h_from_x_eps(
        self,
        x: jax.typing.ArrayLike,
        sample_index: jax.typing.ArrayLike,
        batch_index: jax.typing.ArrayLike,
        extra_eps: float,
        cf_sample: jax.typing.ArrayLike | None = None,
        mc_samples: int = 10,
    ):
        """Compute normalized gene expression from observations using predefined eps"""
        library = 7.0 * jnp.ones_like(
            sample_index
        )  # placeholder, has no effect on the value of h.
        inference_outputs = self.inference(
            x, sample_index, mc_samples=mc_samples, cf_sample=cf_sample, use_mean=False
        )
        generative_inputs = {
            "z": inference_outputs["z_base"] + extra_eps,
            "library": library,
            "batch_index": batch_index,
            "label_index": jnp.zeros([x.shape[0], 1]),
        }
        generative_outputs = self.generative(**generative_inputs)
        return generative_outputs["h"]



from ._base_module import (
    BaseMinifiedModeModuleClass,
    BaseModuleClass,
    JaxBaseModuleClass,
    LossOutput,
    PyroBaseModuleClass,
    TrainStateWithState,
)
from ._decorators import auto_move_data, flax_configure
from ._embedding_mixin import EmbeddingModuleMixin

__all__ = [
    "BaseModuleClass",
    "LossOutput",
    "PyroBaseModuleClass",
    "auto_move_data",
    "flax_configure",
    "JaxBaseModuleClass",
    "TrainStateWithState",
    "BaseMinifiedModeModuleClass",
    "EmbeddingModuleMixin",
]



import warnings

import torch
from torch import nn

from scvi import settings


def one_hot(index: torch.Tensor, n_cat: int) -> torch.Tensor:
    """One hot a tensor of categories."""
    # TODO: remove in 1.3.0
    warnings.warn(
        "The `one_hot` function is deprecated in v1.2 and will be removed in v1.3. "
        "Please use the `one_hot` function in PyTorch instead.",
        DeprecationWarning,
        stacklevel=settings.warnings_stacklevel,
    )
    onehot = torch.zeros(index.size(0), n_cat, device=index.device)
    onehot.scatter_(1, index.type(torch.long), 1)
    return onehot.type(torch.float32)


class ExpActivation(nn.Module):
    def __init__(self):
        super().__init__()

    def forward(self, input):
        return torch.exp(input)



import pytest
import torch

from scvi import REGISTRY_KEYS
from scvi.module import SCANVAE


@pytest.mark.parametrize("n_labels", [1, 2, 10])
def test_scanvae_loss_n_labels(n_labels: int):
    vae = SCANVAE(n_input=100, n_labels=n_labels)

    x = torch.randint(0, 100, (128, 100), dtype=torch.float32)
    batch = torch.randint(0, 5, (128, 1), dtype=torch.long)
    labels = torch.randint(0, 15, (128, 1), dtype=torch.long)
    tensors = {
        REGISTRY_KEYS.X_KEY: x,
        REGISTRY_KEYS.BATCH_KEY: batch,
        REGISTRY_KEYS.LABELS_KEY: labels,
    }

    _ = vae.forward(tensors, inference_kwargs={"n_samples": 1})


@pytest.mark.parametrize("n_samples", [1, 2, 3])
def test_scanvae_loss_n_samples(n_samples: int):
    vae = SCANVAE(n_input=100, n_labels=1)
    vae2 = SCANVAE(n_input=100, n_labels=2)
    vae3 = SCANVAE(n_input=100, n_labels=8)

    x = torch.randint(0, 100, (128, 100), dtype=torch.float32)
    batch = torch.randint(0, 5, (128, 1), dtype=torch.long)
    labels = torch.randint(0, 15, (128, 1), dtype=torch.long)
    tensors = {
        REGISTRY_KEYS.X_KEY: x,
        REGISTRY_KEYS.BATCH_KEY: batch,
        REGISTRY_KEYS.LABELS_KEY: labels,
    }

    _ = vae.forward(tensors, inference_kwargs={"n_samples": n_samples})
    _ = vae2.forward(tensors, inference_kwargs={"n_samples": n_samples})
    _ = vae3.forward(tensors, inference_kwargs={"n_samples": n_samples})



from __future__ import annotations

import logging
from pathlib import Path
from typing import TYPE_CHECKING

import numpy as np
import pandas as pd
import torch

from scvi.data import AnnDataManager
from scvi.data._download import _download
from scvi.data._preprocessing import _dna_to_code
from scvi.data.fields import CategoricalVarField, LayerField, ObsmField
from scvi.dataloaders import DataSplitter
from scvi.external.scbasset._module import REGISTRY_KEYS, ScBassetModule
from scvi.model.base import BaseModelClass
from scvi.train import TrainingPlan, TrainRunner
from scvi.utils import dependencies, setup_anndata_dsp
from scvi.utils._docstrings import devices_dsp

if TYPE_CHECKING:
    from typing import Literal

    from anndata import AnnData

logger = logging.getLogger(__name__)


class SCBASSET(BaseModelClass):
    """``EXPERIMENTAL`` Reimplementation of scBasset :cite:p:`Yuan2022`.

    Performs representation learning of scATAC-seq data. Original implementation:
    https://github.com/calico/scBasset.

    We are working to measure the performance of this model compared to the original.

    Parameters
    ----------
    adata
        single-cell AnnData object that has been registered via
        :meth:`~scvi.external.SCBASSET.setup_anndata`.
    n_bottleneck_layer
        Size of the bottleneck layer
    l2_reg_cell_embedding
        L2 regularization for the cell embedding layer. A value, e.g. 1e-8 can be used to improve
        integration performance.
    **model_kwargs
        Keyword args for :class:`~scvi.external.scbasset.ScBassetModule`

    Examples
    --------
    >>> adata = anndata.read_h5ad(path_to_sc_anndata)
    >>> scvi.data.add_dna_sequence(adata)
    >>> adata = adata.transpose()  # regions by cells
    >>> scvi.external.SCBASSET.setup_anndata(adata, dna_code_key="dna_code")
    >>> model = scvi.external.SCBASSET(adata)
    >>> model.train()
    >>> adata.varm["X_scbasset"] = model.get_latent_representation()

    Notes
    -----
    See further usage examples in the following tutorials:

    1. :doc:`/tutorials/notebooks/atac/scbasset`
    2. :doc:`/tutorials/notebooks/atac/scbasset_batch`
    """

    MOTIF_URLS = {
        "human": (
            "https://storage.googleapis.com/scbasset_tutorial_data/Homo_sapiens_motif_fasta.tar.gz",
            "Homo_sapiens_motif_fasta",
        ),
    }
    DEFAULT_MOTIF_DIR = "./scbasset_motifs/"

    def __init__(
        self,
        adata: AnnData,
        n_bottleneck_layer: int = 32,
        l2_reg_cell_embedding: float = 0.0,
        **model_kwargs,
    ):
        super().__init__(adata)
        self.n_cells = self.summary_stats.n_vars
        self.n_regions = adata.n_obs
        self.n_batch = self.summary_stats.n_batch
        batch_ids = self.adata_manager.get_from_registry(REGISTRY_KEYS.BATCH_KEY)
        self.module = ScBassetModule(
            n_cells=self.n_cells,
            batch_ids=torch.tensor(batch_ids).long() if batch_ids.sum() > 0 else None,
            n_bottleneck_layer=n_bottleneck_layer,
            l2_reg_cell_embedding=l2_reg_cell_embedding,
            **model_kwargs,
        )
        self._model_summary_string = (
            f"ScBasset Model with params: \nn_regions: {self.n_regions}, n_batch: {self.n_batch}, "
            f"n_cells: {self.n_cells}"
        )
        self.init_params_ = self._get_init_params(locals())

    @devices_dsp.dedent
    def train(
        self,
        max_epochs: int = 1000,
        lr: float = 0.01,
        accelerator: str = "auto",
        devices: int | list[int] | str = "auto",
        train_size: float | None = None,
        validation_size: float | None = None,
        shuffle_set_split: bool = True,
        batch_size: int = 128,
        early_stopping: bool = True,
        early_stopping_monitor: str = "auroc_train",
        early_stopping_mode: Literal["min", "max"] = "max",
        early_stopping_min_delta: float = 1e-6,
        datasplitter_kwargs: dict | None = None,
        plan_kwargs: dict | None = None,
        **trainer_kwargs,
    ):
        """Train the model.

        Parameters
        ----------
        max_epochs
            Number of epochs to train for
        lr
            Learning rate for optimization.
        %(param_accelerator)s
        %(param_devices)s
        train_size
            Size of training set in the range [0.0, 1.0].
        validation_size
            Size of the test set. If `None`, defaults to 1 - `train_size`. If
            `train_size + validation_size < 1`, the remaining cells belong to a test set.
        shuffle_set_split
            Whether to shuffle indices before splitting. If `False`, the val, train, and test set
            are split in the sequential order of the data according to `validation_size` and
            `train_size` percentages.
        batch_size
            Minibatch size to use during training.
        early_stopping
            Perform early stopping. Additional arguments can be passed in `**kwargs`.
            See :class:`~scvi.train.Trainer` for further options.
        early_stopping_monitor
            Metric logged during validation set epoch. The available metrics will depend on
            the training plan class used. We list the most common options here in the typing.
        early_stopping_mode
            In 'min' mode, training will stop when the quantity monitored has stopped decreasing
            and in 'max' mode it will stop when the quantity monitored has stopped increasing.
        early_stopping_min_delta
            Minimum change in the monitored quantity to qualify as an improvement,
            i.e. an absolute change of less than min_delta, will count as no improvement.
        datasplitter_kwargs
            Additional keyword arguments passed into :class:`~scvi.dataloaders.DataSplitter`.
        plan_kwargs
            Keyword args for :class:`~scvi.train.TrainingPlan`. Keyword arguments passed to
            `train()` will overwrite values present in `plan_kwargs`, when appropriate.
        **trainer_kwargs
            Other keyword args for :class:`~scvi.train.Trainer`.
        """
        custom_plan_kwargs = {
            "optimizer": "Custom",
            "optimizer_creator": lambda p: torch.optim.Adam(p, lr=lr, betas=(0.95, 0.9995)),
        }
        if plan_kwargs is not None:
            custom_plan_kwargs.update(plan_kwargs)

        datasplitter_kwargs = datasplitter_kwargs or {}

        data_splitter = DataSplitter(
            self.adata_manager,
            train_size=train_size,
            validation_size=validation_size,
            shuffle_set_split=shuffle_set_split,
            batch_size=batch_size,
            # We don't want to dataload the batch ids into the module
            data_and_attributes={
                REGISTRY_KEYS.X_KEY: np.float32,
                REGISTRY_KEYS.DNA_CODE_KEY: np.int64,
            },
            **datasplitter_kwargs,
        )
        training_plan = TrainingPlan(self.module, **custom_plan_kwargs)

        es = {
            "early_stopping": early_stopping,
            "early_stopping_monitor": early_stopping_monitor,
            "early_stopping_mode": early_stopping_mode,
            "early_stopping_min_delta": early_stopping_min_delta,
        }
        for k, v in es.items():
            trainer_kwargs[k] = v if k not in trainer_kwargs.keys() else trainer_kwargs[k]
        runner = TrainRunner(
            self,
            training_plan=training_plan,
            data_splitter=data_splitter,
            max_epochs=max_epochs,
            accelerator=accelerator,
            devices=devices,
            **trainer_kwargs,
        )
        return runner()

    @torch.inference_mode()
    def get_latent_representation(self) -> np.ndarray:
        """Returns the latent representation of the cells.

        Returns
        -------
        latent representation (n_cells, n_latent)
        """
        return self.module.cell_embedding.cpu().numpy().T

    @torch.inference_mode()
    def get_cell_bias(self) -> np.ndarray:
        """Returns the cell-specific bias term.

        Returns
        -------
        bias (n_cells,)
        """
        return self.module.cell_bias.cpu().numpy()

    @classmethod
    def _download_motifs(cls, genome: str, motif_dir: str) -> None:
        """Download a set of motifs injected into peaks"""
        logger.info(f"Downloading motif set to: {motif_dir}")
        # download the motif set
        url_name = cls.MOTIF_URLS.get(genome, None)  # (url, dir_name)
        if url_name is None:
            raise ValueError(f"{genome} is not a supported motif set.")
        _download(url_name[0], save_path=motif_dir, filename=f"{genome}_motifs.tar.gz")
        # untar it
        import tarfile

        def rename_members(tarball):
            """Rename files in the tarball to remove the top level folder"""
            for member in tarball.getmembers():
                if member.path.startswith(url_name[1]):
                    member.path = member.path.replace(url_name[1] + "/", "")
                    yield member

        # importing the "tarfile" module
        tarball = tarfile.open(Path(motif_dir, f"{genome}_motifs.tar.gz"))
        tarball.extractall(path=motif_dir, members=rename_members(tarball))
        tarball.close()

        # `motif_dir` now has `shuffled_peaks_motifs` as a subdir and
        # `shuffled_peaks.fasta` as a root level file.
        logger.info("Download and extraction complete.")
        return

    @dependencies("Bio")
    def _get_motif_library(
        self, tf: str, genome: str = "human", motif_dir: str | None = None
    ) -> tuple[list[str], list[str]]:
        """Load sequences with a TF motif injected from a pre-computed library

        Parameters
        ----------
        tf
            name of the transcription factor motif to load. Must be present in a
            pre-computed library.
        genome
            species name for the motif injection procedure. Currently, only "human"
            is supported.
        motif_dir
            path for the motif library. Will download if not already present.

        Returns
        -------
        motif_seqs
            list of sequences with an injected motif.
        bg_seqs
            dinucleotide shuffled background sequences.
        """
        from Bio import SeqIO

        if motif_dir is None:
            motif_dir = self.DEFAULT_MOTIF_DIR

        # ensure input is a `Path` object
        motif_dir = Path(motif_dir)
        if not Path(motif_dir, "shuffled_peaks.fasta").exists():
            motif_dir.mkdir(exist_ok=True, parents=True)
            self._download_motifs(genome=genome, motif_dir=motif_dir)

        fasta_files = motif_dir.glob("shuffled_peaks_motifs/*.fasta")
        tf_names = [f.stem for f in fasta_files]
        if tf not in tf_names:
            msg = f"{tf} is not found as a motif in the library."
            raise ValueError(msg)

        # load the motifs
        tf_motif_path = Path(motif_dir, "shuffled_peaks_motifs", f"{tf}.fasta")
        motif_seqs = list(SeqIO.parse(tf_motif_path, "fasta"))
        motif_seqs = [str(i.seq) for i in motif_seqs]

        bg_seqs_path = Path(motif_dir, "shuffled_peaks.fasta")
        bg_seqs = list(SeqIO.parse(bg_seqs_path, "fasta"))
        bg_seqs = [str(i.seq) for i in bg_seqs]
        return motif_seqs, bg_seqs

    @torch.inference_mode()
    def get_tf_activity(
        self,
        tf: str,
        genome: str = "human",
        motif_dir: str | None = None,
        lib_size_norm: bool | None = True,
        batch_size: int = 256,
    ) -> np.ndarray:
        """Infer transcription factor activity using a motif injection procedure.

        Parameters
        ----------
        tf
            transcription factor name. must be provided in the relevant motif repository.
        genome
            species name for the motif injection procedure. Currently, only "human"
            is supported.
        motif_dir
            path for the motif library. Will download if not already present.
        lib_size_norm
            normalize accessibility scores for library size by *substracting* the
            cell bias term from each accessibility score prior to comparing motif
            scores to background scores.
        batch_size
            minibatch size for TF activity inference.

        Returns
        -------
        tf_score
            [cells,] TF activity scores.

        Notes
        -----
        scBasset infers TF activities by injecting known TF motifs into a
        shuffled dinucleotide sequence and computing the change in accessibility
        predicted between the injected motif and a randomized background
        sequence. See :cite:p:`Yuan2022` for details. We modeled this function
        off the original implementation in `scbasset`.

        https://github.com/calico/scBasset/blob/9a3294c240d849cdac62682e324bc5f4836bb744/scbasset/utils.py#L453
        """
        # check for a library of FASTA sequences corresponding to motifs and
        # download if none is found
        # `motif_seqs` is a List of str sequences where each char is in "ACTGN".
        # `bg_seqs` is the same, but background sequences rather than motif injected
        motif_seqs, bg_seqs = self._get_motif_library(tf=tf, genome=genome, motif_dir=motif_dir)

        # SCBASSET.module.inference(...) takes `dna_code: torch.Tensor` as input
        # where `dna_code` is [batch_size, seq_length] and each value is [0,1,2,3]
        # where [0: A, 1: C, 2: G, 3: T].
        motif_codes = pd.DataFrame([list(s) for s in motif_seqs]).applymap(_dna_to_code)
        bg_codes = pd.DataFrame([list(s) for s in bg_seqs]).applymap(_dna_to_code)
        # [batch_size, seq_length]
        motif_codes = torch.from_numpy(np.array(motif_codes)).long()
        bg_codes = torch.from_numpy(np.array(bg_codes)).long()

        # NOTE: SCBASSET uses a fixed size of 1344 bp. If motifs from a different source
        # than the above are used, we may need to truncate to match the model size.
        # We should be cautious about doing this, so we throw a warning to the user.
        model_input_size = self.adata_manager.get_from_registry(REGISTRY_KEYS.DNA_CODE_KEY).shape[
            1
        ]
        n_diff = motif_codes.shape[1] - model_input_size
        if n_diff > 0:
            n_cut = n_diff // 2
            logger.warning(
                f"Motif size {motif_codes.shape[1]} != model input size {model_input_size}."
                f" Trimming {n_cut} from each side."
            )
            motif_codes = motif_codes[:, n_cut:-n_cut]
            bg_codes = bg_codes[:, n_cut:-n_cut]
        if n_diff < 0:
            msg = f"Motif sizes {motif_codes.shape[1]} < model size {model_input_size}"
            raise ValueError(msg)

        motif_accessibility = self.module._get_accessibility(
            dna_codes=motif_codes,
            batch_size=batch_size,
        )
        bg_accessibility = self.module._get_accessibility(
            dna_codes=bg_codes,
            batch_size=batch_size,
        )
        # move to CPU
        motif_accessibility = motif_accessibility.detach().cpu()
        bg_accessibility = bg_accessibility.detach().cpu()
        if lib_size_norm:
            # substract the cell bias term so that scores are agnostic to the
            # library size of each observation
            bias = self.module.cell_bias.detach().cpu()
            motif_accessibility = motif_accessibility - bias
            bg_accessibility = bg_accessibility - bias

        # compute the difference in activity between the motif and background
        # sequences
        # after means, arr is activity by cell, shape [cells,]
        motif_activity = motif_accessibility.mean(0) - bg_accessibility.mean(0)
        motif_activity = motif_activity.numpy()
        # z-score the activity
        tf_score = (motif_activity - motif_activity.mean()) / motif_activity.std()
        return tf_score

    @classmethod
    @setup_anndata_dsp.dedent
    def setup_anndata(
        cls,
        adata: AnnData,
        dna_code_key: str,
        layer: str | None = None,
        batch_key: str | None = None,
        **kwargs,
    ):
        """%(summary)s.

        Parameters
        ----------
        %(param_adata)s
        dna_code_key
            Key in `adata.obsm` with dna sequences encoded as integer code.
        %(param_layer)s
        batch_key
            key in `adata.var` for batch information. Categories will automatically be converted
            into integer categories and saved to `adata.var['_scvi_batch']`. If `None`, assigns the
            same batch to all the data.

        Notes
        -----
        The adata object should be in the regions by cells format. This is due to scBasset
        considering regions as observations and cells as variables. This can be simply achieved
        by transposing the data, `bdata = adata.transpose()`.
        """
        setup_method_args = cls._get_setup_method_args(**locals())
        anndata_fields = [
            LayerField(REGISTRY_KEYS.X_KEY, layer, is_count_data=True),
            ObsmField(REGISTRY_KEYS.DNA_CODE_KEY, dna_code_key, is_count_data=True),
            CategoricalVarField(REGISTRY_KEYS.BATCH_KEY, batch_key),
        ]
        adata_manager = AnnDataManager(fields=anndata_fields, setup_method_args=setup_method_args)
        adata_manager.register_fields(adata, **kwargs)
        cls.register_manager(adata_manager)



from collections.abc import Sequence

import torch
import torch.nn as nn


class ConditionalDenseNN(nn.Module):
    """Dense neural network with multiple outputs, optionally conditioned on a context variable.

    (Derived from pyro.nn.dense_nn.ConditionalDenseNN with some modifications [1])

    Parameters
    ----------
    input_dim
        Dimension of the input
    hidden_dims
        Dimensions of the hidden layers (excluding the output layer)
    output_dims
        Dimensions of each output layer
    context_dim
        Dimension of the context input.
    deep_context_injection
        If True, inject the context into every hidden layer.
        If False, only inject the context into the first hidden layer
        (concatenated with the input).
    activation
        Activation function to use between hidden layers (not applied to the outputs).
        Default: torch.nn.ReLU()
    """

    def __init__(
        self,
        input_dim: int,
        hidden_dims: Sequence[int],
        output_dims: Sequence = (1,),
        context_dim: int = 0,
        deep_context_injection: bool = False,
        activation=torch.nn.ReLU(),
    ):
        super().__init__()

        self.input_dim = input_dim
        self.context_dim = context_dim
        self.hidden_dims = hidden_dims
        self.output_dims = output_dims
        self.deep_context_injection = deep_context_injection
        self.n_output_layers = len(self.output_dims)
        self.output_total_dim = sum(self.output_dims)

        # The multiple outputs are computed as a single output layer, and then split
        last_output_end_idx = 0
        self.output_slices = []
        for dim in self.output_dims:
            self.output_slices.append(slice(last_output_end_idx, last_output_end_idx + dim))
            last_output_end_idx += dim

        # Create masked layers
        deep_context_dim = self.context_dim if self.deep_context_injection else 0
        layers = []
        batch_norms = []
        if len(hidden_dims):
            layers.append(torch.nn.Linear(input_dim + context_dim, hidden_dims[0]))
            batch_norms.append(nn.BatchNorm1d(hidden_dims[0]))
            for i in range(1, len(hidden_dims)):
                layers.append(
                    torch.nn.Linear(hidden_dims[i - 1] + deep_context_dim, hidden_dims[i])
                )
                batch_norms.append(nn.BatchNorm1d(hidden_dims[i]))

            layers.append(
                torch.nn.Linear(hidden_dims[-1] + deep_context_dim, self.output_total_dim)
            )
        else:
            layers.append(torch.nn.Linear(input_dim + context_dim, self.output_total_dim))

        self.layers = torch.nn.ModuleList(layers)

        self.activation_fn = activation
        self.batch_norms = torch.nn.ModuleList(batch_norms)

    def forward(self, x, context=None):
        if context is not None:
            # We must be able to broadcast the size of the context over the input
            context = context.expand(x.size()[:-1] + (context.size(-1),))

        h = x
        for i, layer in enumerate(self.layers):
            if self.context_dim > 0 and (self.deep_context_injection or i == 0):
                h = torch.cat([context, h], dim=-1)
            h = layer(h)
            if i < len(self.layers) - 1:
                h = self.batch_norms[i](h)
                h = self.activation_fn(h)

        if self.n_output_layers == 1:
            return h

        h = h.reshape(list(x.size()[:-1]) + [self.output_total_dim])
        return tuple([h[..., s] for s in self.output_slices])



import pytest
import torch

from scvi import REGISTRY_KEYS
from scvi.module import VAE


@pytest.mark.parametrize("n_samples", [1, 2, 3])
@pytest.mark.parametrize("gene_likelihood", ["zinb", "nb", "poisson"])
def test_sample(
    n_samples: int,
    gene_likelihood: str,
    n_input: int = 100,
    batch_size: int = 10,
):
    vae = VAE(n_input=n_input, gene_likelihood=gene_likelihood)
    x = torch.randint(0, 100, (batch_size, n_input), dtype=torch.float32)
    batch = torch.zeros(batch_size, dtype=torch.long)
    labels = torch.zeros(batch_size, dtype=torch.long)
    tensors = {
        REGISTRY_KEYS.X_KEY: x,
        REGISTRY_KEYS.BATCH_KEY: batch,
        REGISTRY_KEYS.LABELS_KEY: labels,
    }

    x_hat = vae.sample(tensors, n_samples=n_samples)
    assert x_hat.dtype == torch.float32
    if n_samples > 1:
        assert x_hat.shape == (batch_size, n_input, n_samples)
    else:
        assert x_hat.shape == (batch_size, n_input)
